import contextlib
import dataclasses
import warnings

from collections.abc import Mapping as C_Mapping
from typing import (
    Callable,
    List,
    Optional,
    Sequence,
    Union,
    Any,
    Dict,
    Iterator,
    MutableMapping,
    Type,
    TYPE_CHECKING,
    overload,
    ClassVar,
    TypeVar,
    Literal,
)
from typing_extensions import dataclass_transform
from functools import wraps

from dlt.common.utils import classlocal

if TYPE_CHECKING:
    TDtcField = dataclasses.Field[Any]
else:
    TDtcField = dataclasses.Field

from dlt.common.typing import (
    AnyType,
    SecretSentinel,
    ConfigValueSentinel,
    TAnyClass,
    Annotated,
    Self,
    extract_inner_type,
    get_type_globals,
    is_annotated,
    is_any_type,
    is_final_type,
    is_optional_type,
    is_subclass,
    is_union_type,
    get_args,
    get_origin,
    resolve_single_annotation,
)
from dlt.common.data_types import py_type_to_sc_type
from dlt.common.configuration.exceptions import (
    ConfigFieldMissingTypeHintException,
    ConfigFieldTypeHintNotSupported,
)


# forward class declaration
_F_BaseConfiguration: Any = type(object)
_F_ContainerInjectableContext: Any = type(object)
_B = TypeVar("_B", bound="BaseConfiguration")


class NotResolved:
    """Used in type annotations to indicate types that should not be resolved."""

    def __init__(self, not_resolved: bool = True):
        self.not_resolved = not_resolved

    def __bool__(self) -> bool:
        return self.not_resolved


def is_hint_not_resolvable(hint: AnyType) -> bool:
    """Checks if hint should NOT be resolved. Final and types annotated like

    >>> Annotated[str, NotResolved()]

    are not resolved.
    """
    if is_final_type(hint):
        return True

    if is_annotated(hint):
        _, *a_m = get_args(hint)
        for annotation in a_m:
            if isinstance(annotation, NotResolved):
                return bool(annotation)
    return False


def is_base_configuration_inner_hint(inner_hint: Type[Any]) -> bool:
    return is_subclass(inner_hint, BaseConfiguration)


def is_context_inner_hint(inner_hint: Type[Any]) -> bool:
    return is_subclass(inner_hint, ContainerInjectableContext)


def is_credentials_inner_hint(inner_hint: Type[Any]) -> bool:
    return is_subclass(inner_hint, CredentialsConfiguration)


def get_config_if_union_hint(hint: Type[Any]) -> Type[Any]:
    if is_union_type(hint):
        return next((t for t in get_args(hint) if is_base_configuration_inner_hint(t)), None)
    return None


def is_valid_hint(hint: Type[Any]) -> bool:
    if get_origin(hint) is ClassVar:
        # class vars are skipped by dataclass
        return True

    if is_hint_not_resolvable(hint):
        # all hints that are not resolved are valid
        return True

    hint = extract_inner_type(hint)
    hint = get_config_if_union_hint(hint) or hint
    hint = get_origin(hint) or hint

    if is_any_type(hint):
        return True
    if is_base_configuration_inner_hint(hint):
        return True
    with contextlib.suppress(TypeError):
        py_type_to_sc_type(hint)
        return True
    return False


def extract_inner_hint(
    hint: Type[Any],
    preserve_new_types: bool = False,
    preserve_literal: bool = False,
    preserve_annotated: bool = False,
) -> Type[Any]:
    # extract hint from Optional / Literal / NewType hints
    inner_hint = extract_inner_type(hint, preserve_new_types, preserve_literal, preserve_annotated)
    # get base configuration from union type
    inner_hint = get_config_if_union_hint(inner_hint) or inner_hint
    # extract origin from generic types (ie List[str] -> List)
    origin = get_origin(inner_hint) or inner_hint
    if preserve_literal and origin is Literal or preserve_annotated and origin is Annotated:
        return inner_hint
    return origin or inner_hint


def is_secret_hint(hint: Type[Any]) -> bool:
    is_secret = False
    if is_annotated(hint):
        _, *a_m = get_args(hint)
        is_secret = SecretSentinel in a_m
    if not is_secret:
        is_secret = is_credentials_inner_hint(hint)
    if not is_secret:
        inner_hint = extract_inner_hint(hint, preserve_annotated=True, preserve_new_types=True)
        # something was encapsulated
        if inner_hint is not hint:
            is_secret = is_secret_hint(inner_hint)
    return is_secret


@overload
def configspec(cls: Type[TAnyClass], init: bool = True) -> Type[TAnyClass]: ...


@overload
def configspec(
    cls: None = ..., init: bool = True
) -> Callable[[Type[TAnyClass]], Type[TAnyClass]]: ...


@dataclass_transform(eq_default=False, field_specifiers=(dataclasses.Field, dataclasses.field))
def configspec(
    cls: Optional[Type[Any]] = None, init: bool = True
) -> Union[Type[TAnyClass], Callable[[Type[TAnyClass]], Type[TAnyClass]]]:
    """Converts (via derivation) any decorated class to a Python dataclass that may be used as a spec to resolve configurations

    __init__ method is synthesized by default. `init` flag is ignored if the decorated class implements custom __init__ as well as
    when any of base classes has no synthesized __init__

    All fields must have default values. This decorator will add `None` default values that miss one.

    In comparison to the Python dataclass, a spec implements full dictionary interface for its attributes, allows instance creation from ie. strings
    or other types (parsing, deserialization) and control over configuration resolution process. See `BaseConfiguration` and CredentialsConfiguration` for
    more information.
    """

    def wrap(cls: Type[TAnyClass]) -> Type[TAnyClass]:
        cls.__hint_resolvers__ = {}  # type: ignore[attr-defined]
        is_context = issubclass(cls, _F_ContainerInjectableContext)

        # if type does not derive from BaseConfiguration then derive it
        with contextlib.suppress(NameError):
            if not issubclass(cls, BaseConfiguration):
                # keep the original module and keep defaults for fields listed in annotations
                fields = {
                    "__module__": cls.__module__,
                    "__annotations__": getattr(cls, "__annotations__", {}),
                }
                for key in fields["__annotations__"].keys():  # type: ignore[union-attr]
                    if key in cls.__dict__:
                        fields[key] = cls.__dict__[key]
                cls = type(cls.__name__, (cls, _F_BaseConfiguration), fields)
        # get all annotations without corresponding attributes and set them to None
        for ann in cls.__annotations__:
            if not hasattr(cls, ann) and not ann.startswith(("__", "_abc_")):
                warnings.warn(
                    f"Missing default value for field {ann} on {cls.__name__}. None assumed. All"
                    " fields in configspec must have defaults."
                )
                setattr(cls, ann, None)
        # get all attributes without corresponding annotations
        globalns = get_type_globals(cls)
        for att_name, att_value in list(cls.__dict__.items()):
            # skip callables, dunder names, class variables and some special names
            if callable(att_value):
                if hint_field_name := getattr(att_value, "__hint_for_field__", None):
                    cls.__hint_resolvers__[hint_field_name] = att_value  # type: ignore[attr-defined]
                    continue
                try:
                    # Allow callable config objects (e.g. Incremental)
                    if not isinstance(att_value, BaseConfiguration):
                        continue
                except NameError:
                    # Dealing with BaseConfiguration itself before it is defined
                    continue
            if not att_name.startswith(("__", "_abc_")) and not isinstance(
                att_value, (staticmethod, classmethod, property)
            ):
                if att_name not in cls.__annotations__:
                    raise ConfigFieldMissingTypeHintException(att_name, cls)
                hint = cls.__annotations__[att_name]
                # resolve the annotation as per PEP 563
                # NOTE: we do not use get_type_hints because at this moment cls is an unknown name
                # (ie. used as decorator and module is being imported)
                hint = resolve_single_annotation(hint, globalns=globalns, raise_on_error=True)
                # context can have any type
                if not is_valid_hint(hint) and not is_context:
                    raise ConfigFieldTypeHintNotSupported(att_name, cls, hint)
                # replace config / secret sentinels
                if isinstance(att_value, ConfigValueSentinel):
                    if is_secret_hint(att_value.default_type) and not is_secret_hint(hint):
                        warnings.warn(
                            f"You indicated {att_name} to be {att_value.default_literal} but type"
                            " hint is not a secret"
                        )
                    if not is_secret_hint(att_value.default_type) and is_secret_hint(hint):
                        warnings.warn(
                            f"You typed {att_name} to be a secret but"
                            f" {att_value.default_literal} indicates it is not"
                        )
                    setattr(cls, att_name, None)

                from types import MappingProxyType

                # make dict immutable: TODO: fix propery
                # somehow default_factory does not work here..
                if isinstance(att_value, dict):
                    setattr(cls, att_name, MappingProxyType(att_value))

                if isinstance(att_value, BaseConfiguration):
                    # Wrap config defaults in default_factory to work around dataclass
                    # blocking mutable defaults
                    def default_factory(att_value=att_value):  # type: ignore[no-untyped-def]
                        return att_value.copy()

                    setattr(cls, att_name, dataclasses.field(default_factory=default_factory))

        # We don't want to overwrite user's __init__ method
        # Create dataclass init only when not defined in the class
        # NOTE: any class without synthesized __init__ breaks the creation chain
        has_default_init = super(cls, cls).__init__ == cls.__init__  # type: ignore[misc]
        base_params = getattr(cls, "__dataclass_params__", None)  # cls.__init__ is object.__init__
        synth_init = init and ((not base_params or base_params.init) and has_default_init)
        if synth_init != init and has_default_init:
            warnings.warn(
                f"__init__ method will not be generated on {cls.__name__} because base class didn't"
                " synthesize __init__. Please correct `init` flag in configspec decorator. You are"
                " probably receiving incorrect __init__ signature for type checking"
            )
        # mark exactly this class as processed by configspec so we can warn the user if it is not done
        # this is not visible in derived classes that must be always decorated with configspec
        cls.__configspec__ = classlocal(True, cls)  # type: ignore[attr-defined]
        # do not generate repr as it may contain secret values
        return dataclasses.dataclass(cls, init=synth_init, eq=False, repr=False)  # type: ignore

    # called with parenthesis
    if cls is None:
        return wrap

    return wrap(cls)


@configspec
class BaseConfiguration(MutableMapping[str, Any]):
    __is_resolved__: bool = dataclasses.field(default=False, init=False, repr=False, compare=False)
    """True when all config fields were resolved and have a specified value type"""
    __exception__: Exception = dataclasses.field(
        default=None, init=False, repr=False, compare=False
    )
    """Holds the exception that prevented the full resolution"""
    __resolved_fields_set__: List[str] = dataclasses.field(
        default=None, init=False, repr=False, compare=False
    )
    """Fields set to non-defaults during resolve, including explicit values"""
    __section__: ClassVar[str] = None
    """Obligatory section used by config providers when searching for keys, always present in the search path"""
    __recommended_sections__: ClassVar[Sequence[str]] = None
    """Recommended sections layout"""
    __config_gen_annotations__: ClassVar[List[str]] = []
    """Additional annotations for config generator, currently holds a list of fields of interest that have defaults"""
    __dataclass_fields__: ClassVar[Dict[str, TDtcField]]
    """Typing for dataclass fields"""
    __hint_resolvers__: ClassVar[Dict[str, Callable[["BaseConfiguration"], Type[Any]]]] = {}

    @classmethod
    def from_init_value(cls: Type[_B], init_value: Any = None) -> _B:
        """Initializes credentials from `init_value`

        Init value may be a native representation of the credentials or a dict. In case of native representation (for example a connection string or JSON with service account credentials)
        a `parse_native_representation` method will be used to parse it. In case of a dict, the credentials object will be updated with key: values of the dict.
        Unexpected values in the dict will be ignored.

        Credentials will be marked as resolved if all required fields are set resolve() method is successful
        """
        # create an instance
        self = cls()
        self._apply_init_value(init_value)
        if not self.is_partial():
            # let it fail gracefully
            with contextlib.suppress(Exception):
                self.resolve()
        return self

    def _apply_init_value(self, init_value: Any = None) -> None:
        if isinstance(init_value, C_Mapping):
            self.update(init_value)
        elif init_value is not None:
            self.parse_native_representation(init_value)

    def parse_native_representation(self, native_value: Any) -> None:
        """Initialize the configuration fields by parsing the `native_value` which should be a native representation of the configuration
        or credentials, for example database connection string or JSON serialized GCP service credentials file.

        Args:
            native_value (Any): A native representation of the configuration

        Raises:
            NotImplementedError: This configuration does not have a native representation
            ValueError: The value provided cannot be parsed as native representation
        """
        raise NotImplementedError()

    def to_native_representation(self) -> Any:
        """Represents the configuration instance in its native form ie. database connection string or JSON serialized GCP service credentials file.

        Raises:
            NotImplementedError: This configuration does not have a native representation

        Returns:
            Any: A native representation of the configuration
        """
        raise NotImplementedError()

    @classmethod
    def _get_resolvable_dataclass_fields(cls) -> Iterator[TDtcField]:
        """Yields all resolvable dataclass fields in the order they should be resolved"""
        # Sort dynamic type hint fields last because they depend on other values
        yield from sorted(
            (f for f in cls.__dataclass_fields__.values() if is_valid_configspec_field(f)),
            key=lambda f: f.name in cls.__hint_resolvers__,
        )

    @classmethod
    def get_resolvable_fields(cls) -> Dict[str, type]:
        """Returns a mapping of fields to their type hints. Dunders should not be resolved and are not returned"""
        globalns = get_type_globals(cls)
        return {
            f.name: resolve_single_annotation(f.type, globalns=globalns)
            for f in cls._get_resolvable_dataclass_fields()
        }

    @classmethod
    def is_field_resolved(cls, value: Any, hint: AnyType) -> bool:
        """Checks if config field with `value` and type `hint` is resolved

        Optional fields are always resolved, required fields must hold value
        and if they are embedded configuration, they must be resolved.
        """
        if not is_optional_type(hint) and (
            value is None
            or ((not value.__is_resolved__) if isinstance(value, BaseConfiguration) else False)
        ):
            return False
        else:
            return True

    def is_resolved(self) -> bool:
        return self.__is_resolved__

    def is_partial(self) -> bool:
        """Returns True when any required resolvable field has its value missing."""
        if self.__is_resolved__:
            return False
        # check if all resolvable fields have value
        return any(
            field
            for field, hint in self.get_resolvable_fields().items()
            if not self.is_field_resolved(getattr(self, field), hint)
        )

    def resolve(self) -> Self:
        self.call_method_in_mro("on_resolved")
        self.__is_resolved__ = True
        return self

    def copy(self: _B) -> _B:
        """Returns recursive copy of the configuration instance. Only embedded configurations
        are copied recursively. In all other cases references are copied.
        """
        new_obj = object.__new__(self.__class__)

        for f in dataclasses.fields(self):
            if not hasattr(self, f.name):
                continue
            v = getattr(self, f.name)
            nv = v.copy() if isinstance(v, BaseConfiguration) else v
            object.__setattr__(new_obj, f.name, nv)

        return new_obj

    def as_dict_nondefault(self) -> Dict[str, Any]:
        """Gets configuration as dictionary containing only values that are non-default"""
        return {
            field.name: self[field.name]
            for field in self._get_resolvable_dataclass_fields()
            if (
                field.default_factory() != self[field.name]
                if field.default_factory != dataclasses.MISSING
                else field.default != self[field.name]
            )
        }

    # implement dictionary-compatible interface on top of dataclass

    def __getitem__(self, __key: str) -> Any:
        if self.__has_attr(__key):
            return getattr(self, __key)
        else:
            raise KeyError(__key)

    def __setitem__(self, __key: str, __value: Any) -> None:
        if self.__has_attr(__key):
            setattr(self, __key, __value)
        else:
            try:
                if not self.__ignore_set_unknown_keys:
                    # assert getattr(self, "__ignore_set_unknown_keys") is not None
                    raise KeyError(__key)
            except AttributeError:
                # __ignore_set_unknown_keys attribute may not be present at the moment of checking, __init__ of BaseConfiguration is not typically called
                raise KeyError(__key)

    def __delitem__(self, __key: str) -> None:
        raise KeyError("Configuration fields cannot be deleted")

    def __iter__(self) -> Iterator[str]:
        """Iterator or valid key names"""
        return map(
            lambda field: field.name,
            filter(lambda val: is_valid_configspec_field(val), self.__dataclass_fields__.values()),
        )

    def __len__(self) -> int:
        return sum(1 for _ in self.__iter__())

    def update(self, other: Any = (), /, **kwds: Any) -> None:
        try:
            self.__ignore_set_unknown_keys = True
            super().update(other, **kwds)
        finally:
            self.__ignore_set_unknown_keys = False

    # helper functions

    def __has_attr(self, __key: str) -> bool:
        return __key in self.__dataclass_fields__ and is_valid_configspec_field(
            self.__dataclass_fields__[__key]
        )

    def call_method_in_mro(config, method_name: str) -> None:
        # python multi-inheritance is cooperative and this would require that all configurations cooperatively
        # call each other class_method_name. this is not at all possible as we do not know which configs in the end will
        # be mixed together.

        # get base classes in order of derivation, base classes first
        mro = reversed(type.mro(type(config)))
        for c in mro:
            # check if this class implements on_resolved (skip pure inheritance to not do double work)
            if method_name in c.__dict__ and callable(getattr(c, method_name)):
                # pass right class instance
                c.__dict__[method_name](config)


_F_BaseConfiguration = BaseConfiguration


def is_valid_configspec_field(field: TDtcField) -> bool:
    return not field.name.startswith("__") and field._field_type is dataclasses._FIELD  # type: ignore


@configspec
class CredentialsConfiguration(BaseConfiguration):
    """Base class for all credentials. Credentials are configurations that may be stored only by providers supporting secrets."""

    __section__: ClassVar[str] = "credentials"

    def to_native_credentials(self) -> Any:
        """Returns native credentials object.

        By default calls `to_native_representation` method.
        """
        return self.to_native_representation()

    def __str__(self) -> str:
        """Get string representation of credentials to be displayed, with all secret parts removed"""
        return super().__str__()


class CredentialsWithDefault:
    """A mixin for credentials that can be instantiated from default ie. from well known env variable with credentials"""

    def has_default_credentials(self) -> bool:
        return hasattr(self, "_default_credentials")

    def _set_default_credentials(self, credentials: Any) -> None:
        self._default_credentials = credentials

    def default_credentials(self) -> Any:
        if self.has_default_credentials():
            return self._default_credentials
        return None


TInjectableContext = TypeVar("TInjectableContext", bound="ContainerInjectableContext")


@configspec
class ContainerInjectableContext(BaseConfiguration):
    """Base class for all configurations that may be injected from a Container. Injectable configuration is called a context"""

    can_create_default: ClassVar[bool] = True
    """If True, `Container` is allowed to create default context instance, if none exists"""
    global_affinity: ClassVar[bool] = False
    """If True, `Container` will create context that will be visible in any thread. If False, per thread context is created"""
    worker_affinity: ClassVar[bool] = False
    """If True, context will be passed to worker processes (e.g., in process pools)."""
    in_container: Annotated[bool, NotResolved()] = dataclasses.field(
        default=False, init=False, repr=False, compare=False
    )
    """Current container, if None then not injected"""
    extras_added: Annotated[bool, NotResolved()] = dataclasses.field(
        default=False, init=False, repr=False, compare=False
    )
    """Tells if extras were already added to this context"""

    def add_extras(self) -> None:
        """Called once after default context was created and added to the container. Benefits mostly the config provider injection context which adds extra providers using the initial ones."""
        pass

    def after_add(self) -> None:
        """Called each time after context is added to container"""

    def before_remove(self) -> None:
        """Called each time before context is removed from container"""

    # context is always resolved
    __is_resolved__ = True


_F_ContainerInjectableContext = ContainerInjectableContext


TSpec = TypeVar("TSpec", bound=BaseConfiguration)
THintResolver = Callable[[TSpec], Type[Any]]


def resolve_type(field_name: str) -> Callable[[THintResolver[TSpec]], THintResolver[TSpec]]:
    def decorator(func: THintResolver[TSpec]) -> THintResolver[TSpec]:
        func.__hint_for_field__ = field_name  # type: ignore[attr-defined]

        @wraps(func)
        def wrapper(self: TSpec) -> Type[Any]:
            return func(self)

        return wrapper

    return decorator



from io import BytesIO
from typing import Tuple, BinaryIO, List
from pathlib import Path
import tarfile
import yaml

from dlt.common.storages.load_package import create_load_id
from dlt.common.utils import digest256_tar_stream

from dlt._workspace.deployment.file_selector import BaseFileSelector
from dlt._workspace.deployment.manifest import (
    TDeploymentFileItem,
    TDeploymentManifest,
    DEPLOYMENT_ENGINE_VERSION,
)

from dlt._workspace._workspace_context import WorkspaceRunContext


DEFAULT_DEPLOYMENT_FILES_FOLDER = "files"
DEFAULT_MANIFEST_FILE_NAME = "manifest.yaml"
DEFAULT_DEPLOYMENT_PACKAGE_LAYOUT = "deployment-{timestamp}.tar.gz"


class PackageBuilder:
    """Builds gzipped deployment package from file selectors"""

    def __init__(self, context: WorkspaceRunContext):
        self.run_context: WorkspaceRunContext = context

    def write_package_to_stream(
        self, file_selector: BaseFileSelector, output_stream: BinaryIO
    ) -> str:
        """Write deployment package to output stream, return content hash"""
        manifest_files: List[TDeploymentFileItem] = []

        # Add files to the archive
        with tarfile.open(fileobj=output_stream, mode="w|gz") as tar:
            for abs_path, rel_path in file_selector:
                # Use POSIX paths for tar archives (cross-platform compatibility)
                posix_path = rel_path.as_posix()
                tar.add(
                    abs_path,
                    arcname=f"{DEFAULT_DEPLOYMENT_FILES_FOLDER}/{posix_path}",
                    recursive=False,
                )
                manifest_files.append(
                    {
                        "relative_path": posix_path,
                        "size_in_bytes": abs_path.stat().st_size,
                    }
                )
            # Create and add manifest with file metadata at the end
            # NOTE: Sort files in manifest because os.scandir(), which the file selector's pathspec.util.iter_tree_files() relies on,
            # yields files in a system-dependent order (https://peps.python.org/pep-0471/#os-scandir).
            manifest: TDeploymentManifest = {
                "engine_version": DEPLOYMENT_ENGINE_VERSION,
                "files": sorted(manifest_files, key=lambda x: x["relative_path"]),
            }
            manifest_yaml = yaml.dump(
                manifest, allow_unicode=True, default_flow_style=False, sort_keys=False
            ).encode("utf-8")
            manifest_info = tarfile.TarInfo(name=DEFAULT_MANIFEST_FILE_NAME)
            manifest_info.size = len(manifest_yaml)
            tar.addfile(manifest_info, BytesIO(manifest_yaml))

        content_hash, _ = digest256_tar_stream(
            output_stream, filter_file_names=lambda x: x != DEFAULT_MANIFEST_FILE_NAME
        )
        return content_hash

    def build_package(self, file_selector: BaseFileSelector) -> Tuple[Path, str]:
        """Create deployment package file, return (path, content_hash)"""
        package_name = DEFAULT_DEPLOYMENT_PACKAGE_LAYOUT.format(timestamp=create_load_id())
        package_path = Path(self.run_context.get_data_entity(package_name))

        with open(package_path, "w+b") as f:
            content_hash = self.write_package_to_stream(file_selector, f)

        return package_path, content_hash



"""Measure the extract, normalize and load progress"""
from typing import Union, Literal

from dlt.common.runtime.collector import (
    TqdmCollector as tqdm,
    LogCollector as log,
    EnlightenCollector as enlighten,
    AliveCollector as alive_progress,
)
from dlt.common.runtime.collector import Collector as _Collector, NULL_COLLECTOR as _NULL_COLLECTOR

TSupportedCollectors = Literal["tqdm", "enlighten", "log", "alive_progress"]
TCollectorArg = Union[_Collector, TSupportedCollectors]


def _from_name(collector: TCollectorArg) -> _Collector:
    """Create default collector by name"""
    if collector is None:
        return _NULL_COLLECTOR

    if isinstance(collector, str):
        if collector == "tqdm":
            return tqdm()
        if collector == "enlighten":
            return enlighten()
        if collector == "log":
            return log()
        if collector == "alive_progress":
            return alive_progress()
        raise ValueError(collector)
    return collector



import contextlib
from collections.abc import Sequence as C_Sequence
from copy import copy
import itertools
from typing import Iterator, List, Dict, Any, Optional
import yaml

from dlt.common import logger
from dlt.common.configuration.container import Container
from dlt.common.configuration.resolve import inject_section
from dlt.common.configuration.specs import ConfigSectionContext, known_sections
from dlt.common.data_writers.writers import EMPTY_DATA_WRITER_METRICS, TDataItemFormat
from dlt.common.pipeline import (
    ExtractDataInfo,
    ExtractInfo,
    ExtractMetrics,
    SupportsPipeline,
    WithStepInfo,
)
from dlt.common.typing import TColumnNames, TLoaderFileFormat
from dlt.common.runtime import signals
from dlt.common.runtime.collector import Collector, NULL_COLLECTOR
from dlt.common.schema import Schema, utils
from dlt.common.schema.typing import (
    TAnySchemaColumns,
    TSchemaContract,
    TTableFormat,
    TWriteDispositionConfig,
)
from dlt.common.storages import NormalizeStorageConfiguration, LoadPackageInfo, SchemaStorage
from dlt.common.storages.load_package import (
    ParsedLoadJobFileName,
    LoadPackageStateInjectableContext,
    TLoadPackageState,
    commit_load_package_state,
)
from dlt.common.utils import (
    get_callable_name,
    get_full_obj_class_name,
    group_dict_of_lists,
    update_dict_nested,
)

from dlt.extract.decorators import (
    _DltSingleSource,
    SourceInjectableContext,
    SourceSchemaInjectableContext,
)
from dlt.extract.exceptions import UnknownSourceReference
from dlt.extract.incremental import IncrementalResourceWrapper
from dlt.extract.items_transform import ItemTransform
from dlt.common.metrics import DataWriterAndCustomMetrics
from dlt.extract.pipe_iterator import PipeIterator
from dlt.extract.source import DltSource
from dlt.extract.reference import SourceReference, SourceFactory
from dlt.extract.resource import DltResource
from dlt.extract.storage import ExtractStorage
from dlt.extract.extractors import ObjectExtractor, ArrowExtractor, Extractor, ModelExtractor
from dlt.extract.state import reset_resource_state
from dlt.extract.utils import get_data_item_format, make_schema_with_default_name


def select_schema(pipeline: SupportsPipeline) -> Schema:
    """Use a clone that will get discarded if extraction fails"""
    # take pipeline schema to make newest version visible to the resources
    if pipeline.default_schema_name:
        name = pipeline.default_schema_name
        # clones with name which will drop previous hashes
        schema_ = pipeline.schemas[name].clone(with_name=name)
        # delete data tables
        schema_.drop_tables(schema_.data_table_names(include_incomplete=True))
    else:
        schema_ = Schema(make_schema_with_default_name(pipeline.pipeline_name))
    return schema_


def data_to_sources(
    data: Any,
    pipeline: SupportsPipeline,
    *,
    schema: Schema = None,
    table_name: str = None,
    parent_table_name: str = None,
    write_disposition: TWriteDispositionConfig = None,
    columns: TAnySchemaColumns = None,
    primary_key: TColumnNames = None,
    table_format: TTableFormat = None,
    schema_contract: TSchemaContract = None,
    loader_file_format: TLoaderFileFormat = None,
) -> List[DltSource]:
    """Creates a list of sources for data items present in `data` and applies specified hints to all resources.

    `data` may be a DltSource, DltResource, a list of those or any other data type accepted by pipeline.run

    The returned list of sources is sorted by schema name.
    """

    def apply_hint_args(resource: DltResource) -> None:
        resource.apply_hints(
            table_name=table_name,
            parent_table_name=parent_table_name,
            write_disposition=write_disposition,
            columns=columns,
            primary_key=primary_key,
            schema_contract=schema_contract,
            table_format=table_format,
            file_format=loader_file_format,
        )

    def apply_settings(source_: DltSource) -> None:
        # apply schema contract settings
        if schema_contract:
            source_.schema_contract = schema_contract

    # a list of sources or a list of resources may be passed as data
    sources: List[DltSource] = []
    resources: Dict[str, List[DltResource]] = {}
    data_resources: List[DltResource] = []

    def append_data(data_item: Any) -> None:
        if isinstance(data_item, DltSource):
            # if schema is explicit then override source schema
            if schema:
                data_item.schema = schema
            sources.append(data_item)
        elif isinstance(data_item, DltResource):
            # many resources with the same name may be present
            r_ = resources.setdefault(data_item.name, [])
            r_.append(data_item)
        elif isinstance(data_item, SourceFactory):
            source = data_item()
            if schema:
                source.schema = schema
            sources.append(source)
        else:
            # iterator/iterable/generator
            # create resource first without table template
            data_resources.append(
                DltResource.from_data(data_item, name=table_name, section=pipeline.pipeline_name)
            )

    if isinstance(data, C_Sequence) and len(data) > 0:
        # if first element is source or resource
        if isinstance(data[0], (DltResource, DltSource)):
            for item in data:
                append_data(item)
        else:
            append_data(data)
    elif isinstance(data, str):
        # TODO: also recognize typical uris like file:// or postgresql://
        # resolve as source reference
        try:
            # get factory and obtain instance of DltSource
            data = SourceReference.from_reference(data)
        except UnknownSourceReference:
            logger.warning(
                "The string you passed as data could not be parsed as valid source reference (ie."
                " `sql_database`)"
            )
        append_data(data)
    else:
        append_data(data)

    # add all appended resource instances in one source
    if resources:
        # decompose into groups so at most single resource with a given name belongs to a group
        for r_ in group_dict_of_lists(resources):
            # wrap into resources with following rules
            # 1. single resource carries its section into source
            # 2. group of resources in the same section carry the section into source
            # 3. resources from different sections: section is pipeline name
            # 4. resources always join default pipeline schema

            r_s = list(r_.values())
            sections = set([r.section for r in r_s])
            section: str = None
            if len(sections) == 1:
                section = iter(sections).__next__()
            section = section or pipeline.pipeline_name
            if len(r_s) == 1:
                # single resource source
                sources.append(_DltSingleSource(schema or select_schema(pipeline), section, r_s))
            else:
                sources.append(DltSource(schema or select_schema(pipeline), section, r_s))

    # add all the appended data-like items in one source
    if data_resources:
        sources.append(
            DltSource(schema or select_schema(pipeline), pipeline.pipeline_name, data_resources)
        )

    # apply hints and settings
    for source in sources:
        apply_settings(source)
        for resource in source.resources.extracted:
            apply_hint_args(resource)

    sources.sort(key=lambda s: s.schema.name)
    return sources


def describe_extract_data(data: Any) -> List[ExtractDataInfo]:
    """Extract source and resource names from data passed to extract"""
    data_info: List[ExtractDataInfo] = []

    def add_item(item: Any) -> bool:
        if isinstance(item, (DltResource, DltSource)):
            # record names of sources/resources
            data_info.append(
                {
                    "name": item.name,
                    "data_type": "resource" if isinstance(item, DltResource) else "source",
                }
            )
            return False
        else:
            # skip None
            if data is not None:
                # any other data type does not have a name - just type
                data_info.append({"name": "", "data_type": type(item).__name__})
            return True

    item: Any = data
    if isinstance(data, C_Sequence) and len(data) > 0:
        for item in data:
            # add_item returns True if non named item was returned. in that case we break
            if add_item(item):
                break
        return data_info

    add_item(item)
    return data_info


class Extract(WithStepInfo[ExtractMetrics, ExtractInfo]):
    original_data: Any
    """Original data from which the extracted DltSource was created. Will be used to describe in extract info"""

    def __init__(
        self,
        schema_storage: SchemaStorage,
        normalize_storage_config: NormalizeStorageConfiguration,
        collector: Collector = NULL_COLLECTOR,
        original_data: Any = None,
    ) -> None:
        """optionally saves originally extracted `original_data` to generate extract info"""
        self.collector = collector
        self.schema_storage = schema_storage
        self.extract_storage = ExtractStorage(normalize_storage_config)
        # TODO: this should be passed together with DltSource to extract()
        self.original_data: Any = original_data
        super().__init__()

    def _compute_metrics(self, load_id: str, source: DltSource) -> ExtractMetrics:
        # map by job id
        job_metrics = {
            ParsedLoadJobFileName.parse(m.file_path): m
            for m in self.extract_storage.closed_files(load_id)
        }
        # aggregate by table name
        table_metrics = {
            table_name: sum(map(lambda pair: pair[1], metrics), EMPTY_DATA_WRITER_METRICS)
            for table_name, metrics in itertools.groupby(
                job_metrics.items(), lambda pair: pair[0].table_name
            )
        }

        # aggregate by resource name
        def _get_all_resource_custom_metrics(resource_name: str) -> Dict[str, Any]:
            all_custom_metrics = source.resources[resource_name].custom_metrics
            for step in source.resources[resource_name]._pipe.steps:
                if isinstance(step, ItemTransform):
                    all_custom_metrics = update_dict_nested(all_custom_metrics, step.custom_metrics)
            return all_custom_metrics

        resource_metrics = {
            resource_name: DataWriterAndCustomMetrics(
                *sum(map(lambda pair: pair[1], metrics), EMPTY_DATA_WRITER_METRICS),
                custom_metrics=_get_all_resource_custom_metrics(resource_name),
            )
            for resource_name, metrics in itertools.groupby(
                table_metrics.items(), lambda pair: source.schema.get_table(pair[0])["resource"]
            )
        }
        # collect resource hints
        clean_hints: Dict[str, Dict[str, Any]] = {}
        for resource in source.selected_resources.values():
            # cleanup the hints
            hints = clean_hints[resource.name] = {}
            resource_hints = copy(resource._hints) or resource.compute_table_schema()
            if resource.incremental and "incremental" not in resource_hints:
                resource_hints["incremental"] = resource.incremental  # type: ignore

            for name, hint in resource_hints.items():
                if hint is None or name in ["validator"]:
                    continue
                if name == "incremental":
                    # represent incremental as dictionary (it derives from BaseConfiguration)
                    if isinstance(hint, IncrementalResourceWrapper):
                        hint = hint.incremental
                    # sometimes internal incremental is not bound
                    if hint:
                        hints[name] = dict(hint)  # type: ignore[call-overload]
                    continue
                if name == "original_columns":
                    # this is original type of the columns ie. Pydantic model
                    hints[name] = get_full_obj_class_name(hint)
                    continue
                if callable(hint):
                    hints[name] = get_callable_name(hint)
                    continue
                if name == "columns":
                    if hint:
                        hints[name] = yaml.dump(
                            hint, allow_unicode=True, default_flow_style=False, sort_keys=False
                        )
                    continue
                hints[name] = hint

        return {
            "started_at": None,
            "finished_at": None,
            "schema_name": source.schema.name,
            "job_metrics": {job.job_id(): metrics for job, metrics in job_metrics.items()},
            "table_metrics": table_metrics,
            "resource_metrics": resource_metrics,
            "dag": source.resources.selected_dag,
            "hints": clean_hints,
        }

    def _write_empty_files(
        self, source: DltSource, extractors: Dict[TDataItemFormat, Extractor]
    ) -> None:
        schema = source.schema
        json_extractor = extractors["object"]
        resources_with_items = set().union(*[e.resources_with_items for e in extractors.values()])
        # find REPLACE resources that did not yield any pipe items and create empty jobs for them
        # NOTE: do not include tables that have never seen data
        data_tables = {t["name"]: t for t in schema.data_tables(seen_data_only=True)}
        tables_by_resources = utils.group_tables_by_resource(data_tables)
        for resource in source.resources.selected.values():
            if resource.write_disposition != "replace" or resource.name in resources_with_items:
                continue
            if resource.name not in tables_by_resources:
                continue
            for table in tables_by_resources[resource.name]:
                # we only need to write empty files for the root tables
                if not utils.is_nested_table(table):
                    json_extractor.write_empty_items_file(table["name"])

        # collect resources that received empty materialized lists and had no items
        resources_with_empty = (
            set()
            .union(*[e.resources_with_empty for e in extractors.values()])
            .difference(resources_with_items)
        )
        # get all possible tables
        data_tables = {t["name"]: t for t in schema.data_tables()}
        tables_by_resources = utils.group_tables_by_resource(data_tables)
        for resource_name in resources_with_empty:
            if resource := source.resources.selected.get(resource_name):
                if tables := tables_by_resources.get("resource_name"):
                    # write empty tables
                    for table in tables:
                        # we only need to write empty files for the root tables
                        if not utils.is_nested_table(table):
                            json_extractor.write_empty_items_file(table["name"])
                else:
                    table_name = json_extractor._get_static_table_name(resource, None)
                    if table_name:
                        json_extractor.write_empty_items_file(table_name)

    def _extract_single_source(
        self,
        load_id: str,
        source: DltSource,
        *,
        max_parallel_items: int,
        workers: int,
    ) -> None:
        schema = source.schema
        collector = self.collector
        extractors: Dict[TDataItemFormat, Extractor] = {
            "object": ObjectExtractor(
                load_id, self.extract_storage.item_storages["object"], schema, collector=collector
            ),
            "arrow": ArrowExtractor(
                load_id, self.extract_storage.item_storages["arrow"], schema, collector=collector
            ),
            "model": ModelExtractor(
                load_id, self.extract_storage.item_storages["model"], schema, collector=collector
            ),
        }
        # make sure we close storage on exception
        with collector(f"Extract {source.name}"):
            with self.manage_writers(load_id, source):
                # yield from all selected pipes
                with PipeIterator.from_pipes(
                    source.resources.selected_pipes,
                    max_parallel_items=max_parallel_items,
                    workers=workers,
                ) as pipes:
                    left_gens = total_gens = len(pipes._sources)
                    collector.update("Resources", 0, total_gens)
                    for pipe_item in pipes:
                        curr_gens = len(pipes._sources)
                        if left_gens > curr_gens:
                            delta = left_gens - curr_gens
                            left_gens -= delta
                            collector.update("Resources", delta)

                        # kill extraction if signalled
                        signals.raise_if_signalled()

                        resource = source.resources.with_pipe(pipe_item.pipe)
                        item_format = get_data_item_format(pipe_item.item)
                        extractors[item_format].write_items(
                            resource, pipe_item.item, pipe_item.meta
                        )

                    self._write_empty_files(source, extractors)
                    if left_gens > 0:
                        # go to 100%
                        collector.update("Resources", left_gens)

    @contextlib.contextmanager
    def manage_writers(self, load_id: str, source: DltSource) -> Iterator[ExtractStorage]:
        self._step_info_start_load_id(load_id)
        # self.current_source = source
        try:
            yield self.extract_storage
        except Exception:
            # kill writers without flushing the content
            self.extract_storage.close_writers(load_id, skip_flush=True)
            raise
        else:
            self.extract_storage.close_writers(load_id)
        finally:
            # gather metrics when storage is closed
            self.gather_metrics(load_id, source)

    def gather_metrics(self, load_id: str, source: DltSource) -> None:
        # NOTE: there may be more than one extract run per load id: ie. the resource and then dlt state
        # so metrics are immutable here and will be appended
        self._step_info_update_metrics(
            load_id, self._compute_metrics(load_id, source), immutable=True
        )
        self._step_info_complete_load_id(load_id)
        # remove the metrics of files processed in this extract run
        self.extract_storage.remove_closed_files(load_id)

    def extract(
        self,
        source: DltSource,
        max_parallel_items: int,
        workers: int,
        load_package_state_update: Optional[TLoadPackageState] = None,
    ) -> str:
        # generate load package to be able to commit all the sources together later
        load_id = self.extract_storage.create_load_package(
            source.schema, reuse_exiting_package=True
        )
        with (
            Container().injectable_context(SourceSchemaInjectableContext(source.schema)),
            Container().injectable_context(SourceInjectableContext(source)),
            Container().injectable_context(
                LoadPackageStateInjectableContext(
                    load_id=load_id, storage=self.extract_storage.new_packages
                )
            ) as load_package,
        ):
            # inject the config section with the current source name
            with inject_section(
                ConfigSectionContext(
                    sections=(known_sections.SOURCES, source.section, source.name),
                    source_state_key=source.name,
                )
            ):
                if load_package_state_update:
                    load_package.state.update(load_package_state_update)

                # reset resource states, the `extracted` list contains all the explicit resources and all their parents
                for resource in source.resources.extracted:
                    if resource.write_disposition == "replace":
                        reset_resource_state(resource.name)

                self._extract_single_source(
                    load_id,
                    source,
                    max_parallel_items=max_parallel_items,
                    workers=workers,
                )
                commit_load_package_state()
        return load_id

    def commit_packages(self) -> None:
        """Commits all extracted packages to normalize storage"""
        # commit load packages
        for load_id, metrics in self._load_id_metrics.items():
            self.extract_storage.commit_new_load_package(
                load_id, self.schema_storage[metrics[0]["schema_name"]]
            )
        # all load ids got processed, cleanup empty folder
        self.extract_storage.delete_empty_extract_folder()

    def get_step_info(self, pipeline: SupportsPipeline) -> ExtractInfo:
        load_ids = list(self._load_id_metrics.keys())
        load_packages: List[LoadPackageInfo] = []
        metrics: Dict[str, List[ExtractMetrics]] = {}
        for load_id in self._load_id_metrics.keys():
            load_package = self.extract_storage.get_load_package_info(load_id)
            load_packages.append(load_package)
            metrics[load_id] = self._step_info_metrics(load_id)
        return ExtractInfo(
            pipeline,
            metrics,
            describe_extract_data(self.original_data),
            load_ids,
            load_packages,
            pipeline.first_run,
        )



from typing import Optional, List, Dict, Union, Literal
from dlt.common.schema.typing import TColumnSchema, TColumnHint

TDatabricksColumnHint = Union[TColumnHint, Literal["foreign_key"]]


class TDatabricksColumnSchema(TColumnSchema, total=False):
    column_comment: Optional[str]
    column_tags: Optional[List[Union[str, Dict[str, str]]]]


TDatabricksTableSchemaColumns = Dict[str, TDatabricksColumnSchema]



from typing import Any, Iterable, List, Sequence
import textwrap

from dlt.common.exceptions import DltException, TerminalException, TransientException
from dlt.common.reflection.exceptions import ReferenceImportError
from dlt.common.reflection.ref import ImportTrace


class DestinationException(DltException):
    pass


class UnknownDestinationModule(ReferenceImportError, DestinationException, KeyError):
    def __init__(
        self,
        ref: str,
        qualified_refs: Sequence[str],
        traces: Sequence[ImportTrace],
        destination_type: str = None,
        named_dest_attempted: bool = False,
    ) -> None:
        self.ref = ref
        self.qualified_refs = qualified_refs
        self.destination_type = destination_type
        self.named_dest_attempted = named_dest_attempted
        super().__init__(traces=traces)

    def __str__(self) -> str:
        msg = ""
        if "." in self.ref:
            msg += f"Destination module `{self.ref}` is not registered."
        else:
            if self.named_dest_attempted:
                msg += (
                    f"Destination '{self.ref}' was first attempted to be resolved as a named"
                    " destination with a configured type. "
                )
                if self.destination_type:
                    msg += (
                        f"However, the configured destination type '{self.destination_type}' is not"
                        " valid. Set a valid destination type. "
                    )
                else:
                    msg += (
                        "However, no destination type was configured. "
                        "If your destination is a named destination, "
                        "set a valid destination type either as an environment variable:\n\n"
                    )
                    msg += textwrap.indent(
                        f"DESTINATION__{self.ref.upper()}__DESTINATION_TYPE=duckdb\n", "  "
                    )
                    msg += "\nor in your configuration files:\n\n"
                    msg += textwrap.indent(
                        f'[destination.{self.ref}]\ndestination_type="duckdb"\n\n', "  "
                    )

                msg += (
                    f"Since no{' valid' if self.destination_type else ''} destination type was"
                    f" found, dlt also tried to resolve '{self.ref}' as a standard destination."
                    " However, "
                )

            msg += (
                f"{'d' if self.named_dest_attempted else 'D'}estination `{self.ref}` is not one of"
                " the standard dlt destinations."
            )

        if len(self.qualified_refs) == 1 and self.qualified_refs[0] == self.ref:
            pass
        else:
            msg += (
                " Following fully qualified refs were tried in the registry:\n\t%s\n"
                % "\n\t".join(self.qualified_refs)
            )
        if self.traces:
            msg += super().__str__()
        return msg


class InvalidDestinationReference(DestinationException):
    def __init__(self, refs: Any) -> None:
        self.refs = refs
        msg = (
            f"None of supplied destination refs: `{refs}` can be found in registry or imported as"
            " Python type."
        )
        super().__init__(msg)


class DestinationTerminalException(DestinationException, TerminalException):
    pass


class DestinationUndefinedEntity(DestinationTerminalException):
    pass


class DestinationTransientException(DestinationException, TransientException):
    pass


class DestinationLoadingViaStagingNotSupported(DestinationTerminalException):
    def __init__(self, destination: str) -> None:
        self.destination = destination
        super().__init__(f"`{destination=:}` does not support loading via staging.")


class DestinationLoadingWithoutStagingNotSupported(DestinationTerminalException):
    def __init__(self, destination: str) -> None:
        self.destination = destination
        super().__init__(f"`{destination=:}` does not support loading without staging.")


class DestinationNoStagingMode(DestinationTerminalException):
    def __init__(self, destination: str) -> None:
        self.destination = destination
        super().__init__(f"`{destination=:}` cannot be used as a staging")


class DestinationIncompatibleLoaderFileFormatException(DestinationTerminalException):
    def __init__(
        self, destination: str, staging: str, file_format: str, supported_formats: Iterable[str]
    ) -> None:
        self.destination = destination
        self.staging = staging
        self.file_format = file_format
        self.supported_formats = supported_formats
        supported_formats_str = ", ".join(supported_formats)
        if self.staging:
            if not supported_formats:
                msg = (
                    f"`{staging=:}` cannot be used with `{destination=:}` because they"
                    " have no file formats in common."
                )
            else:
                msg = (
                    f"Unsupported `{file_format=:}` for `{destination=:}` with `{staging=:}` "
                    f"Supported formats: `{supported_formats_str}`"
                )
        else:
            msg = (
                f"Unsupported `{file_format=:}` for `{destination=:}`Supported formats:"
                f" {supported_formats_str}. If {destination} supports loading data via staging"
                " bucket, more formats may be available."
            )
        super().__init__(msg)


class IdentifierTooLongException(DestinationTerminalException):
    def __init__(
        self,
        destination_name: str,
        identifier_type: str,
        identifier_name: str,
        max_identifier_length: int,
    ) -> None:
        self.destination_name = destination_name
        self.identifier_type = identifier_type
        self.identifier_name = identifier_name
        self.max_identifier_length = max_identifier_length
        super().__init__(
            f"The length of {identifier_type} {identifier_name} exceeds"
            f" {max_identifier_length} allowed for {destination_name}"
        )


class UnsupportedDataType(DestinationTerminalException):
    def __init__(
        self,
        destination_type: str,
        table_name: str,
        column: str,
        data_type: str,
        file_format: str,
        available_in_formats: Sequence[str],
        more_info: str,
    ) -> None:
        self.destination_type = destination_type
        self.table_name = table_name
        self.column = column
        self.data_type = data_type
        self.file_format = file_format
        self.available_in_formats = available_in_formats
        self.more_info = more_info
        msg = (
            f"Destination `{destination_type}` cannot load `{data_type=:}` from"
            f" `{file_format=:}` files. The affected table is `{table_name}` column `{column}`."
        )
        if available_in_formats:
            msg += f" Note: `{data_type=:}` can be loaded from format(s): `{available_in_formats}`."
        else:
            msg += f" No available file formats for this destination support `{data_type=:}`"
        if more_info:
            msg += " More info: " + more_info
        super().__init__(msg)


class WithJobError:
    """A mixin for exceptions raised on failed jobs"""

    load_id: str
    job_id: str
    failed_message: str


class DestinationHasFailedJobs(DestinationTerminalException):
    def __init__(self, destination_name: str, load_id: str, failed_jobs: List[Any]) -> None:
        self.destination_name = destination_name
        self.load_id = load_id
        self.failed_jobs = failed_jobs
        super().__init__(
            f"Destination `{destination_name}` has failed jobs in load package `{load_id}`"
        )


class DestinationSchemaTampered(DestinationTerminalException):
    def __init__(self, schema_name: str, version_hash: str, stored_version_hash: str) -> None:
        self.version_hash = version_hash
        self.stored_version_hash = stored_version_hash
        super().__init__(
            f"Schema `{schema_name}` content was changed - by a loader or by destination code -"
            " from the moment it was retrieved by load package. Such schema cannot reliably be"
            " updated nor saved. If you are using destination client directly, without storing"
            " schema in load package, you should first save it into schema storage. You can also"
            " use schema._bump_version() in test code to remove modified flag.Version hash:"
            f" `{version_hash=:}` != {stored_version_hash=:}"
        )


class DestinationCapabilitiesException(DestinationException):
    pass


class WriteDispositionNotSupported(DestinationCapabilitiesException):
    def __init__(self, write_disposition: str, message: str):
        self.write_disposition = write_disposition
        super().__init__(message)


class DestinationInvalidFileFormat(DestinationTerminalException):
    def __init__(
        self, destination_type: str, file_format: str, file_name: str, message: str
    ) -> None:
        self.destination_type = destination_type
        self.file_format = file_format
        self.message = message
        super().__init__(
            f"Destination `{destination_type}` cannot process file `{file_name=:}` with"
            f" {file_format=:}: {message}"
        )


class TableFormatNotSupported(DestinationTerminalException):
    def __init__(self, table_format: str, table_name: str, message: str) -> None:
        self.table_format = table_format
        self.table_name = table_name
        self.message = message
        super().__init__(
            f"Table `{table_name}` has unsupported table format `{table_format}`: {message}"
        )


class OpenTableFormatNotSupported(DestinationTerminalException):
    def __init__(self, table_format: str, table_name: str, detected_table_format: str):
        self.table_format = table_format
        self.table_name = table_name
        self.detected_table_format = detected_table_format
        if detected_table_format:
            msg = (
                f"Table `{table_name}` is stored as format `{detected_table_format}` while"
                f" `{table_format=:}` was requested"
            )
        else:
            msg = f"Table `{table_name}` is not stored in any known open table format."
        super().__init__(msg)


class OpenTableCatalogNotSupported(DestinationTerminalException):
    def __init__(self, table_format: str, destination_type: str):
        self.table_format = table_format
        self.destination_type = destination_type
        super().__init__(f"Catalog not supported for `{table_format=:}` in `{destination_type=:}`")


class SqlClientNotAvailable(DestinationTerminalException):
    def __init__(self, entity_type: str, entity_name: str, destination_name: str) -> None:
        super().__init__(
            f"SQL Client not available for {entity_type} `{entity_name}` in destination"
            f" `{destination_name}`",
        )


class DatasetNotAvailable(DestinationTerminalException):
    def __init__(self, destination_name: str) -> None:
        super().__init__(f"Destination `{destination_name}` does not support datasets.")


class OpenTableClientNotAvailable(DestinationTerminalException):
    def __init__(self, dataset_name: str, destination_name: str) -> None:
        super().__init__(
            f"Open table client not available for destination `{destination_name}` in dataset"
            f" `{dataset_name}`",
        )


class FSClientNotAvailable(DestinationTerminalException):
    def __init__(self, pipeline_name: str, destination_name: str) -> None:
        super().__init__(
            f"Filesystem Client not available for destination `{destination_name}` in pipeline"
            f" `{pipeline_name}`",
        )



from typing import Any
from dlt.common.exceptions import MissingDependencyException

try:
    import pandas
    from pandas import DataFrame
except ModuleNotFoundError:
    raise MissingDependencyException("dlt Pandas Helpers", ["pandas"])


def pandas_to_arrow(df: pandas.DataFrame, preserve_index: bool = False) -> Any:
    """Converts pandas to arrow or raises an exception if pyarrow is not installed"""
    from dlt.common.libs.pyarrow import pyarrow as pa

    # NOTE: None preserves named indexes but ignores unnamed
    return pa.Table.from_pandas(df, preserve_index=preserve_index)



from typing import Literal, Optional

from dlt.common.configuration import configspec
from dlt.common.configuration.specs import BaseConfiguration

TPoolType = Literal["process", "thread", "none"]


@configspec
class PoolRunnerConfiguration(BaseConfiguration):
    pool_type: TPoolType = None
    """type of pool to run, must be set in derived configs"""
    start_method: Optional[str] = None
    """start method for the pool (typically process). None is system default"""
    workers: Optional[int] = None
    """# how many threads/processes in the pool"""
    run_sleep: float = 0.1
    """how long to sleep between runs with workload, seconds"""



from typing import Optional

DLT_PIPELINE_COMMAND_DOCS_URL = (
    "https://dlthub.com/docs/reference/command-line-interface#dlt-pipeline"
)
DLT_INIT_DOCS_URL = "https://dlthub.com/docs/reference/command-line-interface#dlt-init"
DLT_TELEMETRY_DOCS_URL = "https://dlthub.com/docs/reference/telemetry"
DLT_DEPLOY_DOCS_URL = "https://dlthub.com/docs/walkthroughs/deploy-a-pipeline"
DLT_AI_DOCS_URL = "https://dlthub.com/docs/dlt-ecosystem/llm-tooling/llm-native-workflow"

DEFAULT_AI_WORKBENCH_REPO = "https://github.com/dlt-hub/dlthub-ai-workbench.git"
DEFAULT_AI_WORKBENCH_BRANCH: Optional[str] = None  # must be None for dlt release!
DEFAULT_AI_WORKBENCH_LICENSE_URL = (
    "https://github.com/dlt-hub/dlthub-ai-workbench/blob/master/LICENSE"
)



import dataclasses
from typing import Final, Optional

from dlt.common.typing import TSecretStrValue
from dlt.common.configuration import configspec
from dlt.common.utils import digest128

from dlt.destinations.impl.postgres.configuration import (
    PostgresCredentials,
    PostgresClientConfiguration,
)


@configspec(init=False)
class RedshiftCredentials(PostgresCredentials):
    port: int = 5439
    password: TSecretStrValue = None
    username: str = None
    host: str = None
    client_encoding: Optional[str] = "utf-8"


@configspec
class RedshiftClientConfiguration(PostgresClientConfiguration):
    destination_type: Final[str] = dataclasses.field(default="redshift", init=False, repr=False, compare=False)  # type: ignore
    credentials: RedshiftCredentials = None

    staging_iam_role: Optional[str] = None
    has_case_sensitive_identifiers: bool = False

    def fingerprint(self) -> str:
        """Returns a fingerprint of host part of a connection string"""
        if self.credentials and self.credentials.host:
            return digest128(self.credentials.host)
        return ""



import os
import ast
import shutil
import warnings
from importlib.metadata import Distribution
from typing import Dict, Optional, Sequence, Tuple

from dlt.common.libs import git
from dlt.common.configuration.specs import known_sections
from dlt.common.configuration.providers import (
    SECRETS_TOML,
    ConfigTomlProvider,
    SecretsTomlProvider,
)
from dlt.common.pipeline import get_dlt_repos_dir
from dlt.version import DLT_PKG_NAME, __version__
from dlt.common.destination import Destination
from dlt.common.reflection.utils import rewrite_python_script
from dlt.common.runtime import run_context
from dlt.common.schema.utils import is_valid_schema_name
from dlt.common.schema.exceptions import InvalidSchemaName
from dlt.common.storages.file_storage import FileStorage

import dlt.destinations
from dlt.sources import SourceReference
import dlt.reflection.names as n
from dlt.reflection.script_inspector import import_pipeline_script

from dlt._workspace.cli import echo as fmt, _pipeline_files as files_ops, source_detection, utils
from dlt._workspace.cli.config_toml_writer import WritableConfigValue, write_values
from dlt._workspace.cli._pipeline_files import (
    TEMPLATE_FILES,
    SOURCES_MODULE_NAME,
    SINGLE_FILE_TEMPLATE_MODULE_NAME,
    SourceConfiguration,
    TVerifiedSourceFileEntry,
    TVerifiedSourceFileIndex,
)
from dlt._workspace.cli.exceptions import CliCommandException, CliCommandInnerException
from dlt._workspace.cli._urls import DLT_INIT_DOCS_URL, DLT_AI_DOCS_URL  # noqa: F401


def list_sources_command(repo_location: str, branch: str = None) -> None:
    fmt.echo("---")
    fmt.echo("Available dlt core sources:")
    fmt.echo("---")
    core_sources = _list_core_sources()
    for source_name, source_configuration in core_sources.items():
        msg = "%s: %s" % (fmt.bold(source_name), source_configuration.doc)
        fmt.echo(msg)

    fmt.echo("---")
    fmt.echo("Available dlt single file templates:")
    fmt.echo("---")
    template_sources = _list_template_sources()
    for source_name, source_configuration in template_sources.items():
        msg = "%s: %s" % (fmt.bold(source_name), source_configuration.doc)
        fmt.echo(msg)

    fmt.echo("---")
    fmt.echo("Available verified sources:")
    fmt.echo("---")
    for source_name, source_configuration in _list_verified_sources(repo_location, branch).items():
        reqs = source_configuration.requirements
        dlt_req_string = str(reqs.dlt_requirement_base)
        msg = "%s: " % (fmt.bold(source_name))
        if source_name in core_sources.keys():
            msg += "(Deprecated since dlt 1.0.0 in favor of core source of the same name) "
        msg += source_configuration.doc
        if not reqs.is_installed_dlt_compatible():
            msg += fmt.warning_style(" [needs update: %s]" % (dlt_req_string))

        fmt.echo(msg)


def list_destinations_command() -> None:
    fmt.echo("---")
    fmt.echo("Available dlt core destinations:")
    fmt.echo("---")
    core_destinations = _list_core_destinations()
    for destination_name in core_destinations:
        msg = "%s" % fmt.bold(destination_name)
        fmt.echo(msg)


def init_command(
    source_name: str,
    destination_type: str,
    repo_location: str,
    branch: str = None,
    eject_source: bool = False,
    dry_run: bool = False,
    add_example_pipeline_script: bool = True,
) -> Tuple[Dict[str, str], files_ops.TSourceType]:
    # detect and warn on deprecated dlthub: pattern
    display_source_name: Optional[str] = None
    if source_name.startswith("dlthub:"):
        display_source_name = source_name[7:]
        fmt.warning(
            "The `dlthub:<source>` syntax is deprecated. User dltHub AI Workbench instead:\n"
            "dlt ai init\n"
            "will get you started. See %s for more details."
            % fmt.bold(DLT_AI_DOCS_URL)
        )
        source_name = "context_rest_api"

    run_ctx = run_context.active()
    destination_storage_path = run_ctx.run_dir
    settings_dir = run_ctx.settings_dir
    sources_dir = run_ctx.get_run_entity("sources")

    return init_pipeline_at_destination(
        source_name,
        destination_type,
        repo_location,
        branch,
        eject_source,
        dry_run,
        add_example_pipeline_script,
        destination_storage_path,
        settings_dir,
        sources_dir,
        display_source_name=display_source_name,
    )


def init_pipeline_at_destination(
    source_name: str,
    destination_type: str,
    repo_location: str,
    branch: str = None,
    eject_source: bool = False,
    dry_run: bool = False,
    add_example_pipeline_script: bool = True,
    destination_storage_path: str = None,
    settings_dir: str = None,
    sources_dir: str = None,
    target_dependency_system: str = None,
    display_source_name: str = None,
) -> Tuple[Optional[Dict[str, str]], files_ops.TSourceType]:
    """
    Initializes a pipeline at the specified destination by setting up the required files, configurations, and dependencies.

    This function handles the discovery of the source type (template, core, or verified), prepares the destination storage,
    resolves conflicts for existing files, and generates necessary configuration files (e.g., `config.toml`, `secrets.toml`).
    It also validates compatibility with the installed dlt version and optionally creates an example pipeline script,
    that loads data from the source to a specifiable destination.

    Args:
        - source_name (str): The name of the source to initialize.
        - destination_type (str): The type of destination (e.g., "bigquery", "redshift").
        - repo_location (str): The location of the verified sources repository.
        - branch (str, optional): The branch of the repository to use. Defaults to None.
        - eject_source (bool, optional): Whether to eject the source code. Defaults to False.
        - dry_run (bool, optional): If True, no files are modified, and the changes are returned as a preview. Defaults to False.
        - skip_example_pipeline_script (bool, optional): If True, skips creating the example pipeline script. Defaults to False.
        - destination_storage_path (str, optional): Path to the destination storage. Defaults to None.
        - settings_dir (str, optional): Path to the settings directory. Defaults to None.
        - sources_dir (str, optional): Path to the sources directory. Defaults to None.
        - target_dependency_system (str, optional): Additional context used to adjust the welcome message. Valid options are
            `requirements.txt` or `pyproject.toml`, Default to None, in which case it will be determined based on the destination storage.

    Returns:
        Tuple[Dict[str, str], Dict[str, WritableConfigValue], Dict[str, WritableConfigValue], files_ops.TSourceType]:
            A tuple containing:
            - A dictionary of copied files (destination path -> source path).
            - The type of the source (e.g., "template", "core", "verified").
            - Name of the selected ide for dlthub sources (defaults to "cursor")
    """
    # validate the user-facing name (display_source_name for dlthub: sources)
    name_to_validate = display_source_name or source_name
    # source and destination names are used as Python identifiers in generated code
    if not is_valid_schema_name(name_to_validate):
        fmt.error(
            "Source name %s is not a valid Python identifier. Use snake_case names"
            " containing only lowercase letters, numbers and underscores (max %d"
            " characters)."
            % (fmt.bold(name_to_validate), InvalidSchemaName.MAXIMUM_SCHEMA_NAME_LENGTH)
        )
        raise CliCommandException()
    # try to import the destination and get config spec
    if destination_type:
        destination_reference = Destination.from_reference(destination_type)
        destination_spec = destination_reference.spec

    # lookup core storages
    core_sources_storage = files_ops.get_core_sources_storage()
    templates_storage = files_ops.get_single_file_templates_storage()

    # discover type of source
    source_type: files_ops.TSourceType = "template"
    if source_name in files_ops.get_sources_names(core_sources_storage, source_type="core"):
        source_type = "core"
    # skip verified sources lookup for dlthub: sources (always use template)
    elif not display_source_name:
        verified_sources_storage = _clone_and_get_verified_sources_storage(repo_location, branch)
        if source_name in files_ops.get_sources_names(
            verified_sources_storage, source_type="verified"
        ):
            source_type = "verified"

    # prepare destination storage
    dest_storage = FileStorage(destination_storage_path)
    if not dest_storage.has_folder(settings_dir):
        dest_storage.create_folder(settings_dir)
    # get local index of verified source files
    local_index = files_ops.load_verified_sources_local_index(source_name)
    # folder deleted at dest - full refresh
    if not dest_storage.has_folder(source_name):
        local_index["files"] = {}
    # is update or new source
    is_new_source = len(local_index["files"]) == 0

    # look for existing source
    source_configuration: SourceConfiguration = None
    remote_index: TVerifiedSourceFileIndex = None
    remote_modified: Dict[str, TVerifiedSourceFileEntry] = {}
    remote_deleted: Dict[str, TVerifiedSourceFileEntry] = {}

    if source_type == "verified":
        # get pipeline files
        source_configuration = files_ops.get_verified_source_configuration(
            verified_sources_storage, source_name
        )
        # get file index from remote verified source files being copied
        remote_index = files_ops.get_remote_source_index(
            source_configuration.storage.storage_path,
            source_configuration.files,
            source_configuration.requirements.dlt_version_constraint(),
        )
        # diff local and remote index to get modified and deleted files
        remote_new, remote_modified, remote_deleted = files_ops.gen_index_diff(
            local_index, remote_index
        )
        # find files that are modified locally
        conflict_modified, conflict_deleted = files_ops.find_conflict_files(
            local_index, remote_new, remote_modified, remote_deleted, dest_storage
        )
        # add new to modified
        remote_modified.update(remote_new)
        if conflict_modified or conflict_deleted:
            # select source files that can be copied/updated
            _, remote_modified, remote_deleted = _select_source_files(
                source_name, remote_modified, remote_deleted, conflict_modified, conflict_deleted
            )
        if not remote_deleted and not remote_modified:
            fmt.echo("No files to update, exiting")
            return None, source_type

        if remote_index["is_dirty"]:
            fmt.warning(
                f"The verified sources repository is dirty. {source_name} source files may"
                " not update correctly in the future."
            )

    else:
        if source_type == "core":
            source_configuration = files_ops.get_core_source_configuration(
                core_sources_storage, source_name, eject_source
            )
            dist = Distribution.from_name(DLT_PKG_NAME)
            extras = dist.metadata.get_all("Provides-Extra") or []

            # Match the extra name to the source name
            canonical_source_name = source_name.replace("_", "-").lower()

            if canonical_source_name in extras:
                source_configuration.requirements.update_dlt_extras(canonical_source_name)

            #  create remote modified index to copy files when ejecting
            remote_modified = {file_name: None for file_name in source_configuration.files}
        else:
            # is single file template source
            source_configuration = files_ops.get_template_configuration(
                templates_storage, source_name, display_source_name or source_name
            )

        if dest_storage.has_file(source_configuration.dest_pipeline_script):
            fmt.warning(
                "Pipeline script %s already exists, exiting"
                % source_configuration.dest_pipeline_script
            )
            return None, source_type

    # add .dlt/*.toml files to be copied
    # source_configuration.files.extend(
    #     [run_ctx.get_setting(CONFIG_TOML), run_ctx.get_setting(SECRETS_TOML)]
    # )

    # add dlt extras line to requirements
    if destination_type:
        source_configuration.requirements.update_dlt_extras(destination_type)

    # Check compatibility with installed dlt
    if not source_configuration.requirements.is_installed_dlt_compatible():
        msg = (
            "This pipeline requires a newer version of dlt than your installed version"
            f" ({source_configuration.requirements.current_dlt_version()}). Pipeline requires"
            f" '{source_configuration.requirements.dlt_requirement_base}'"
        )
        fmt.warning(msg)
        if not fmt.confirm(
            "Would you like to continue anyway? (you can update dlt after this step)", default=True
        ):
            fmt.echo(
                "You can update dlt with: pip3 install -U"
                f' "{source_configuration.requirements.dlt_requirement_base}"'
            )
            return None, source_type

    # read module source and parse it
    visitor = utils.parse_init_script(
        "init",
        source_configuration.storage.load(source_configuration.src_pipeline_script),
        source_configuration.src_pipeline_script,
    )
    if visitor.is_destination_imported:
        raise CliCommandInnerException(
            "init",
            f"The pipeline script {source_configuration.src_pipeline_script} imports a destination"
            " from dlt.destinations. You should specify destinations by name when calling"
            " dlt.pipeline or dlt.run in init scripts.",
        )
    if n.PIPELINE not in visitor.known_calls:
        raise CliCommandInnerException(
            "init",
            f"The pipeline script {source_configuration.src_pipeline_script} does not seem to"
            " initialize a pipeline with dlt.pipeline. Please initialize pipeline explicitly in"
            " your init scripts.",
        )

    # find all arguments in all calls to replace
    transformed_nodes = source_detection.find_call_arguments_to_replace(
        visitor,
        [
            ("destination", destination_type or "duckdb", True),
        ],
        source_configuration.src_pipeline_script,
    )

    # inspect the script to populate source references
    # suppress warnings emitted during import (e.g. psutil missing from LogCollector)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        if source_configuration.source_type != "core":
            import_pipeline_script(
                source_configuration.storage.storage_path,
                source_configuration.storage.to_relative_path(
                    source_configuration.src_pipeline_script
                ),
                ignore_missing_imports=True,
            )
        else:
            # core sources are imported directly from the pipeline script
            # which is in the _workspace module
            import_pipeline_script(
                os.path.dirname(source_configuration.src_pipeline_script),
                os.path.basename(source_configuration.src_pipeline_script),
                ignore_missing_imports=True,
            )

    # detect all the required secrets and configs that should go into tomls files
    if source_configuration.source_type == "template":
        # replace destination, pipeline_name and dataset_name in templates
        # template sources are always in module starting with "pipeline"
        # for templates, place config and secrets into top level section
        required_secrets, required_config, checked_sources = source_detection.detect_source_configs(
            SourceReference.SOURCES, source_configuration.source_module_prefix, ()
        )
        # dlthub: sources get extra AST transforms: rename pipeline, dataset, sources
        if display_source_name:
            transformed_nodes = source_detection.find_call_arguments_to_replace(
                visitor,
                [
                    ("destination", destination_type or "duckdb", True),
                    ("pipeline_name", display_source_name + "_pipeline", True),
                    ("dataset_name", display_source_name + "_data", False),
                ],
                source_configuration.src_pipeline_script,
            )
            for source_q_name, source_config in checked_sources.items():
                if source_q_name not in visitor.known_sources_resources:
                    raise CliCommandInnerException(
                        "init",
                        "The pipeline script %s imports a source/resource %s from"
                        " section %s. In init scripts you must declare all sources"
                        " and resources in single file. Known names are %s."
                        % (
                            source_configuration.src_pipeline_script,
                            source_config.name,
                            source_config.section,
                            list(visitor.known_sources_resources.keys()),
                        ),
                    )
            transformed_nodes.extend(
                source_detection.find_source_calls_to_replace(visitor, display_source_name)
            )
    else:
        # pipeline sources are in module with name starting from {pipeline_name}
        # for verified pipelines place in the specific source section
        required_secrets, required_config, checked_sources = source_detection.detect_source_configs(
            SourceReference.SOURCES,
            source_configuration.source_module_prefix,
            (known_sections.SOURCES, source_name),
        )
        if len(checked_sources) == 0:
            raise CliCommandInnerException(
                "init",
                f"The pipeline script {source_configuration.src_pipeline_script} is not creating or"
                " importing any sources or resources. Exiting...",
            )

    # add destination spec to required secrets
    if destination_type or add_example_pipeline_script:
        required_secrets["destinations:" + destination_type] = WritableConfigValue(
            destination_type, destination_spec, None, ("destination",)
        )
    # add the global telemetry to required config
    required_config["runtime.dlthub_telemetry"] = WritableConfigValue(
        "dlthub_telemetry", bool, utils.get_telemetry_status(), ("runtime",)
    )

    # modify the script
    script_lines = rewrite_python_script(visitor.source_lines, transformed_nodes)
    dest_script_source = "".join(script_lines)
    # validate by parsing
    ast.parse(source=dest_script_source)

    # ask for confirmation
    if is_new_source:
        if source_configuration.source_type == "core":
            fmt.echo(
                "Creating a new pipeline with the dlt core source %s (%s)"
                % (fmt.bold(source_name), source_configuration.doc)
            )
            if eject_source:
                fmt.echo(
                    "NOTE: Source code of %s will be ejected. Remember to modify the pipeline "
                    "example script to import the ejected source." % (fmt.bold(source_name))
                )
            else:
                fmt.echo(
                    "NOTE: Beginning with dlt 1.0.0, the source %s will no longer be copied from"
                    " the verified sources repo but imported from dlt.sources. You can provide the"
                    " --eject flag to revert to the old behavior." % (fmt.bold(source_name))
                )
        elif source_configuration.source_type == "verified":
            new_entity_type = "a new pipeline with" if destination_type else ""
            fmt.echo(
                "Creating and configuring %s the verified source %s (%s)"
                % (new_entity_type, fmt.bold(source_name), source_configuration.doc)
            )
        else:
            if source_configuration.is_default_template:
                fmt.echo(
                    "NOTE: Could not find a dlt source or template wih the name %s. Selecting the"
                    " default template." % (fmt.bold(source_name))
                )
                fmt.echo(
                    "NOTE: In case you did not want to use the default template, run 'dlt init -l'"
                    " to see all available sources and templates."
                )
            fmt.echo(
                "Creating and configuring a new pipeline with the dlt core template %s (%s)"
                % (
                    fmt.bold(source_configuration.dest_pipeline_script),
                    source_configuration.doc,
                )
            )

        if not fmt.confirm("Do you want to proceed?", default=True):
            raise CliCommandInnerException("init", "Aborted")

    dependency_system = target_dependency_system or _get_dependency_system(dest_storage)
    _welcome_message(
        source_name,
        destination_type,
        source_configuration,
        dependency_system,
        is_new_source,
        add_example_pipeline_script,
    )

    # copy files at the very end
    copy_files = []
    # copy template files
    for file_name in TEMPLATE_FILES:
        dest_path = dest_storage.make_full_path(file_name)
        if templates_storage.has_file(file_name):
            if dest_storage.has_file(dest_path):
                # do not overwrite any init files
                continue
            copy_files.append((templates_storage.make_full_path(file_name), dest_path))

    # only those that were modified should be copied from verified sources
    for file_name in remote_modified:
        copy_files.append(
            (
                source_configuration.storage.make_full_path(file_name),
                # copy into where "sources" reside in run context, being root dir by default
                dest_storage.make_full_path(os.path.join(sources_dir, file_name)),
            )
        )
    # if dry-run, do not actually modify storage, just return file content
    pipeline_script_target_path = dest_storage.make_full_path(
        os.path.join(sources_dir, source_configuration.dest_pipeline_script)
    )
    if dry_run:
        files_to_create: Dict[str, str] = {}
        for source_path, dest_path in copy_files:
            try:
                files_to_create[dest_path] = dest_storage.load(source_path)
            except UnicodeDecodeError:
                fmt.warning(
                    f"File {source_path} was skipped not a text file. It will not be copied to"
                    f" {dest_path}"
                )
        if add_example_pipeline_script:
            files_to_create[pipeline_script_target_path] = dest_script_source
        # todo: handle remote index changes?
        return files_to_create, source_type

    # modify storage
    else:
        for src_path, dest_path in copy_files:
            os.makedirs(os.path.dirname(dest_path), exist_ok=True)
            shutil.copy2(src_path, dest_path)
        if remote_index:
            # delete files
            for file_name in remote_deleted:
                if dest_storage.has_file(file_name):
                    dest_storage.delete(file_name)
            files_ops.save_verified_source_local_index(
                source_name, remote_index, remote_modified, remote_deleted
            )
        # create example script
        if (
            not dest_storage.has_file(source_configuration.dest_pipeline_script)
            and add_example_pipeline_script
        ):
            dest_storage.save(pipeline_script_target_path, dest_script_source)

        # generate tomls with comments
        secrets_prov = SecretsTomlProvider(settings_dir)
        write_values(secrets_prov._config_toml, required_secrets.values(), overwrite_existing=False)

        config_prov = ConfigTomlProvider(settings_dir)
        write_values(config_prov._config_toml, required_config.values(), overwrite_existing=False)

        # write toml files
        secrets_prov.write_toml()
        config_prov.write_toml()

        # if there's no dependency system write the requirements file
        if dependency_system is None:
            requirements_txt = "\n".join(source_configuration.requirements.compiled())
            dest_storage.save(utils.REQUIREMENTS_TXT, requirements_txt)

        copied_files: Dict[str, str] = {dest_path: src_path for src_path, dest_path in copy_files}
        return copied_files, source_type


def _clone_and_get_verified_sources_storage(repo_location: str, branch: str = None) -> FileStorage:
    """Clone and get FileStorage for verified sources templates"""

    fmt.echo("Looking up verified sources at %s..." % fmt.bold(repo_location))
    clone_storage = git.get_fresh_repo_files(repo_location, get_dlt_repos_dir(), branch=branch)
    # copy dlt source files from here
    return FileStorage(clone_storage.make_full_path(SOURCES_MODULE_NAME))


def _select_source_files(
    source_name: str,
    remote_modified: Dict[str, TVerifiedSourceFileEntry],
    remote_deleted: Dict[str, TVerifiedSourceFileEntry],
    conflict_modified: Sequence[str],
    conflict_deleted: Sequence[str],
) -> Tuple[str, Dict[str, TVerifiedSourceFileEntry], Dict[str, TVerifiedSourceFileEntry]]:
    # some files were changed and cannot be updated (or are created without index)
    fmt.echo(
        "Existing files for %s source were changed and cannot be automatically updated"
        % fmt.bold(source_name)
    )
    if conflict_modified:
        fmt.echo(
            "Following files are MODIFIED locally and CONFLICT with incoming changes: %s"
            % fmt.bold(", ".join(conflict_modified))
        )
    if conflict_deleted:
        fmt.echo(
            "Following files are DELETED locally and CONFLICT with incoming changes: %s"
            % fmt.bold(", ".join(conflict_deleted))
        )
    can_update_files = set(remote_modified.keys()) - set(conflict_modified)
    can_delete_files = set(remote_deleted.keys()) - set(conflict_deleted)
    if len(can_update_files) > 0 or len(can_delete_files) > 0:
        if len(can_update_files) > 0:
            fmt.echo(
                "Following files can be automatically UPDATED: %s"
                % fmt.bold(", ".join(can_update_files))
            )
        if len(can_delete_files) > 0:
            fmt.echo(
                "Following files can be automatically DELETED: %s"
                % fmt.bold(", ".join(can_delete_files))
            )
        prompt = (
            "Should incoming changes be Skipped, Applied (local changes will be lost) or Merged (%s"
            " UPDATED | %s DELETED | all local changes remain)?"
            % (fmt.bold(",".join(can_update_files)), fmt.bold(",".join(can_delete_files)))
        )
        choices = "sam"
    else:
        prompt = "Should incoming changes be Skipped or Applied?"
        choices = "sa"
    # Skip / Apply / Merge
    resolution = fmt.prompt(prompt, choices, default="s")
    if resolution == "s":
        # do not copy nor delete any files
        fmt.echo("Skipping all incoming changes. No local files were modified.")
        remote_modified.clear()
        remote_deleted.clear()
    elif resolution == "m":
        # update what we can
        fmt.echo("Merging the incoming changes. No files with local changes were modified.")
        remote_modified = {n: e for n, e in remote_modified.items() if n in can_update_files}
        remote_deleted = {n: e for n, e in remote_deleted.items() if n in can_delete_files}
    else:
        # fully overwrite, leave all files to be copied
        fmt.echo("Applying all incoming changes to local files.")

    return resolution, remote_modified, remote_deleted


def _welcome_message(
    source_name: str,
    destination_type: str,
    source_configuration: SourceConfiguration,
    dependency_system: str,
    is_new_source: bool,
    added_pipeline_script: bool = True,
) -> None:
    new_entity_type = "pipeline" if destination_type else "source"
    fmt.echo()
    if source_configuration.source_type in ["template", "core"]:
        fmt.echo(
            "Your new %s %s is ready to be customized!" % (new_entity_type, fmt.bold(source_name))
        )
        if added_pipeline_script:
            fmt.echo(
                "* Review and change how dlt loads your data in %s"
                % fmt.bold(source_configuration.dest_pipeline_script)
            )
    else:
        if is_new_source:
            fmt.echo("Verified source %s was added to your project!" % fmt.bold(source_name))
            if added_pipeline_script:
                fmt.echo(
                    "* See the usage examples and code snippets to copy from %s"
                    % fmt.bold(source_configuration.dest_pipeline_script)
                )
        else:
            fmt.echo(
                "Verified source %s was updated to the newest version!" % fmt.bold(source_name)
            )

    if is_new_source:
        destination_str = " for %s" % fmt.bold(destination_type) if destination_type else ""
        fmt.echo(
            "* Add credentials%s and other secrets to %s"
            % (destination_str, fmt.bold(utils.make_dlt_settings_path(SECRETS_TOML)))
        )

    if destination_type == "destination":
        fmt.echo(
            "* You have selected the custom destination as your pipelines destination. Please refer"
            " to our docs at https://dlthub.com/docs/dlt-ecosystem/destinations/destination on how"
            " to add a destination function that will consume your data."
        )

    if dependency_system:
        fmt.echo("* Add the required dependencies to %s:" % fmt.bold(dependency_system))
        compiled_requirements = source_configuration.requirements.compiled()
        for dep in compiled_requirements:
            fmt.echo("  " + fmt.bold(dep))
        if destination_type:
            fmt.echo(
                "  If the dlt dependency is already added, make sure you install the extra for %s"
                " to it"
                % fmt.bold(destination_type)
            )
        if dependency_system == utils.REQUIREMENTS_TXT:
            qs = "' '"
            fmt.echo(
                "  To install with pip: %s"
                % fmt.bold(f"pip3 install '{qs.join(compiled_requirements)}'")
            )
        elif dependency_system == utils.PYPROJECT_TOML:
            fmt.echo("  If you are using poetry you may issue the following command:")
            fmt.echo(fmt.bold("  poetry add %s -E %s" % (DLT_PKG_NAME, destination_type)))
        fmt.echo()
    else:
        fmt.echo(
            "* %s was created. Install it with:\npip3 install -r %s"
            % (fmt.bold(utils.REQUIREMENTS_TXT), utils.REQUIREMENTS_TXT)
        )

    if is_new_source and new_entity_type == "pipeline":
        fmt.echo(
            "* Read %s for more information"
            % fmt.bold("https://dlthub.com/docs/walkthroughs/create-a-pipeline")
        )
    else:
        fmt.echo(
            "* Read %s for more information"
            % fmt.bold("https://dlthub.com/docs/walkthroughs/add-a-verified-source")
        )


def _get_dependency_system(dest_storage: FileStorage) -> str:
    if dest_storage.has_file(utils.PYPROJECT_TOML):
        return utils.PYPROJECT_TOML
    elif dest_storage.has_file(utils.REQUIREMENTS_TXT):
        return utils.REQUIREMENTS_TXT
    else:
        return None


def _list_template_sources() -> Dict[str, SourceConfiguration]:
    template_storage = files_ops.get_single_file_templates_storage()
    sources: Dict[str, SourceConfiguration] = {}
    for source_name in files_ops.get_sources_names(template_storage, source_type="template"):
        sources[source_name] = files_ops.get_template_configuration(
            template_storage, source_name, source_name
        )
    return sources


def _list_core_sources() -> Dict[str, SourceConfiguration]:
    core_sources_storage = files_ops.get_core_sources_storage()

    sources: Dict[str, SourceConfiguration] = {}
    for source_name in files_ops.get_sources_names(core_sources_storage, source_type="core"):
        sources[source_name] = files_ops.get_core_source_configuration(
            core_sources_storage, source_name, eject_source=False
        )
    return sources


def _list_verified_sources(
    repo_location: str, branch: str = None
) -> Dict[str, SourceConfiguration]:
    verified_sources_storage = _clone_and_get_verified_sources_storage(repo_location, branch)

    sources: Dict[str, SourceConfiguration] = {}
    for source_name in files_ops.get_sources_names(
        verified_sources_storage, source_type="verified"
    ):
        try:
            sources[source_name] = files_ops.get_verified_source_configuration(
                verified_sources_storage, source_name
            )
        except Exception as ex:
            fmt.warning(f"Verified source {source_name} not available: {ex}")

    return sources


def _list_core_destinations() -> list[str]:
    return dlt.destinations.__all__


@utils.track_command("init", False, "source_name", "destination_type")
def init_command_wrapper(
    source_name: str,
    destination_type: str,
    repo_location: str,
    branch: str,
    eject_source: bool = False,
) -> None:
    init_command(
        source_name,
        destination_type,
        repo_location,
        branch,
        eject_source,
    )


@utils.track_command("list_sources", False)
def list_sources_command_wrapper(repo_location: str, branch: str) -> None:
    list_sources_command(repo_location, branch)


@utils.track_command("list_destinations", False)
def list_destinations_command_wrapper() -> None:
    list_destinations_command()



from dlt.common.destination.exceptions import DestinationTerminalException


class InvalidInMemoryDuckdbCredentials(DestinationTerminalException):
    def __init__(self) -> None:
        super().__init__(
            "To use in-memory instance of duckdb, "
            "please instantiate it first and then pass to destination factory\n"
            '\nconn = duckdb.connect(":memory:")\n'
            'dlt.pipeline(pipeline_name="...", destination=dlt.destinations.duckdb(conn)'
        )


class IcebergViewException(DestinationTerminalException):
    def __init__(self, dbapi_exception: Exception, note: str) -> None:
        self.note = note
        super().__init__(dbapi_exception)

    def __str__(self) -> str:
        msg = super().__str__()
        return f"{self.note}\n{msg}"



"""Cli for rendering markdown docs"""

import argparse
import textwrap
import os
import re
from typing import List

import dlt._workspace.cli.echo as fmt

HEADER = """---
title: Command Line Interface
description: Command line interface (CLI) full reference of dlt
keywords: [command line interface, cli, dlt init]
---


# Command Line Interface Reference

<!-- this page is fully generated from the argparse object of dlt, run make update-cli-docs to update it -->

This page contains all commands available in the dlt CLI and is generated
automatically from the fully populated python argparse object of dlt.
:::note
Flags and positional commands are inherited from the parent command. Position within the command string
is important. For example if you want to enable debug mode on the pipeline command, you need to add the
debug flag to the base dlt command:

```sh
dlt --debug pipeline
```

Adding the flag after the pipeline keyword will not work.
:::

"""

# Developer NOTE: This generation is based on parsing the output of the help string in argparse.
# It works very well at the moment, but there may be cases where it will break due to unanticipated
# argparse help output. The cleaner solution would be to implement an argparse help formatter,
# but this would require more work and also some not very nice parsing.


class _WidthFormatter(argparse.RawTextHelpFormatter):
    def __init__(self, prog: str) -> None:
        super().__init__(prog, width=99999, max_help_position=99999)


def render_argparse_markdown(
    name: str,
    parser: argparse.ArgumentParser,
    /,
    *,
    header: str = HEADER,
    commands: List[str] = None,
) -> str:
    def get_parser_help_recursive(
        parser: argparse.ArgumentParser,
        cmd: str = "",
        parent: str = "",
        nesting: int = 0,
        help_string: str = None,
        commands: List[str] = None,
    ) -> str:
        markdown = ""

        # Prevent wrapping in help output for better parseability
        parser.formatter_class = _WidthFormatter
        if parser.description:
            # remove markdown from description
            parser.description = parser.description.markup  # type: ignore
        help_output = parser.format_help()
        sections = help_output.split("\n\n")

        def _text_wrap_line(line: str, indent: int = 4) -> str:
            return "\n".join(
                textwrap.wrap(
                    line,
                    width=80,
                    break_on_hyphens=False,
                    subsequent_indent=" " * indent,
                    break_long_words=False,
                )
            )

        # extract usage section denoted by "usage: "
        usage = sections[0]
        usage = usage.replace("usage: ", "")
        usage = _text_wrap_line(usage)

        # get description or use help passed down from choices
        help_string = help_string or ""
        help_string = help_string.strip().strip(" .") + "."
        description = parser.description or help_string or ""
        description = description.strip().strip(" .") + "."

        if nesting == 0:
            help_string = description
            description = None

        if not parser.description:
            fmt.warning(f"No description found for {cmd}, please consider providing one.")

        inherits_from = ""
        if parent:
            parent_slug = parent.lower().replace(" ", "-")
            inherits_from = f"Inherits arguments from [`{parent}`](#{parent_slug})."

        # extract all other sections
        # here we remove excess information and style the args nicely
        # into markdown lists
        extracted_sections = []
        for section in sections[1:]:
            section_lines = section.splitlines()

            # detect full sections
            if not section_lines[0].endswith(":"):
                continue

            if len(section_lines[0]) > 30:
                continue

            # remove first line with header and empty lines
            header = section_lines[0].replace(":", "")

            if header.lower() not in ["available subcommands", "positional arguments", "options"]:
                fmt.warning(f"Skipping unknown section {header} of {cmd}.")
                continue

            section_lines = section_lines[1:]
            section_lines = [line for line in section_lines if line]
            section_lines = [line for line in section_lines if not line.strip().startswith("{")]
            section = textwrap.dedent(os.linesep.join(section_lines))

            # NOTE: this is based on convention to name the subcommands section
            # "available subcommands"
            is_subcommands_list = "subcommands" in header

            # split section and remove lines starting with {
            section_lines = section.splitlines()
            # make markdown list of args
            section = ""
            for line in section_lines:
                line = line.strip()
                line_elements = re.split(r"\s{2,}|\n+", line)
                arg_title = line_elements[0]
                arg_help = line_elements[1] if len(line_elements) > 1 else ""
                if not arg_help:
                    fmt.warning(
                        f"Missing helpstring for argument '{arg_title}' in section '{header}' of"
                        f" command '{cmd}'."
                    )

                quoted_title = ""
                if is_subcommands_list:
                    # skip unwanted commands
                    if not commands or arg_title in commands:
                        full_command = f"{cmd} {arg_title}"
                        anchor_slug = full_command.lower().replace(" ", "-")
                        quoted_title = f"[`{arg_title}`](#{anchor_slug})"
                else:
                    quoted_title = f"`{arg_title}`"
                if quoted_title:
                    section += f"* {quoted_title} - {arg_help.capitalize()}\n"

            extracted_sections.append({"header": header.capitalize(), "section": section})

        heading = "##" if nesting < 2 else "###"
        markdown += f"{heading} `{cmd}`\n\n"
        if help_string:
            markdown += f"{help_string}\n\n"
        markdown += "**Usage**\n"
        markdown += "```sh\n"
        markdown += f"{usage}\n"
        markdown += "```\n\n"

        if description:
            markdown += "**Description**\n\n"
            markdown += f"{description}\n\n"

        markdown += "<details>\n\n"
        markdown += "<summary>Show Arguments and Options</summary>\n\n"

        if inherits_from:
            markdown += f"{inherits_from}\n\n"

        for es in extracted_sections:
            markdown += f"**{es['header']}**\n"
            markdown += f"{es['section']}\n"

        markdown += "</details>\n\n"

        # traverse the subparsers and forward help strings to the recursive function
        commands_found = []
        for action in parser._actions:
            if isinstance(action, argparse._SubParsersAction):
                for subaction in action._get_subactions():
                    subparser = action._name_parser_map[subaction.dest]
                    # skip unwanted commands
                    if commands and subaction.dest not in commands:
                        continue
                    commands_found.append(subaction.dest)
                    assert (
                        subaction.help
                    ), f"Subparser help string of {subaction.dest} is empty, please provide one."
                    markdown += get_parser_help_recursive(
                        subparser,
                        f"{cmd} {subaction.dest}",
                        parent=cmd,
                        nesting=nesting + 1,
                        help_string=subaction.help,
                        commands=None,
                    )
        if commands and set(commands_found) != set(commands):
            raise RuntimeError(
                f"Following commands were expected:: {commands} but found only {commands_found}."
                " Typically this means that dlthub was not installed or workspace context was "
                " not found."
            )

        return markdown

    markdown = get_parser_help_recursive(parser, name, commands=commands)

    return header + markdown



"""SQL query execution, row counts, and load history retrieval."""

import traceback
from typing import Any, Dict, List, Optional, Tuple, Union

from dlt._workspace.helpers.dashboard.typing import TLoadItem

import marimo as mo

import dlt
import pyarrow
import sqlglot

from dlt.common.destination.exceptions import SqlClientNotAvailable
from dlt.destinations.exceptions import DatabaseUndefinedRelation, DestinationUndefinedEntity
from dlt.pipeline.exceptions import PipelineConfigMissing

from dlt._workspace.helpers.dashboard import strings
from dlt._workspace.helpers.dashboard.config import DashboardConfiguration
from dlt._workspace.helpers.dashboard.utils.formatters import (
    format_exception_message,
    humanize_datetime_values,
)
from dlt._workspace.helpers.dashboard.utils.schema import get_schema_by_version
from dlt._workspace.helpers.dashboard.utils.ui import BoundedDict


def clear_query_cache() -> None:
    """Clear the cached query results."""
    _query_result_cache.clear()


def update_query_history(
    current_history: Dict[str, int],
    query: str,
    row_count: int,
) -> Dict[str, int]:
    """Update query history dict, placing most recent query first."""
    current_history.pop(query, None)
    return {query: row_count, **current_history}


def get_default_query_for_table(
    pipeline: dlt.Pipeline, schema_name: str, table_name: str, limit: bool
) -> Tuple[str, str, str]:
    """Build a default SELECT query for a table. Returns (sql_query, error_message, traceback)."""
    try:
        _dataset = pipeline.dataset(schema=schema_name)
        _sql_query = (
            _dataset.table(table_name)
            .limit(1000 if limit else None)
            .to_sql(pretty=True, _raw_query=True)
        )
        return _sql_query, None, None
    except Exception as exc:
        return "", format_exception_message(exc), traceback.format_exc()


def get_example_query_for_dataset(pipeline: dlt.Pipeline, schema_name: str) -> Tuple[str, str, str]:
    """Return an example query for the first data table in the schema.

    Returns (sql_query, error_message, traceback).
    """
    schema = pipeline.schemas.get(schema_name)
    if schema and (tables := schema.data_tables()):
        return get_default_query_for_table(pipeline, schema_name, tables[0]["name"], True)
    return "", "Schema does not contain any tables.", None


def get_query_result(pipeline: dlt.Pipeline, query: str) -> Tuple[pyarrow.Table, str, str]:
    """Get the result of a query.

    Parses the query to ensure it is valid SQL before sending it to the destination.
    """
    try:
        sqlglot.parse_one(
            query,
            dialect=pipeline.destination.capabilities().sqlglot_dialect,
        )
        return (
            _execute_query_cached(pipeline.pipeline_name, pipeline.dataset_name, query, pipeline),
            None,
            None,
        )
    except Exception as exc:
        return pyarrow.table({}), format_exception_message(exc), traceback.format_exc()


def _execute_query_cached(
    pipeline_name: str, dataset_name: str, query: str, pipeline: dlt.Pipeline
) -> pyarrow.Table:
    """Execute a query, returning cached result if the same pipeline/query was seen before.

    The cache is keyed on (pipeline_name, dataset_name, query) strings so that
    results survive pipeline re-attachment across refreshes.  The pipeline object
    is passed through for execution but is *not* part of the cache key.
    """
    cache_key = (pipeline_name, dataset_name, query)
    if cache_key not in _query_result_cache:
        _query_result_cache[cache_key] = pipeline.dataset()(query, _execute_raw_query=True).arrow()
    return _query_result_cache[cache_key]


_query_result_cache: BoundedDict = BoundedDict(64)


def get_row_counts(
    pipeline: dlt.Pipeline, selected_schema_name: str = None, load_id: str = None
) -> Dict[str, int]:
    """Get the row counts for a pipeline as a dict of {table_name: count}."""
    row_counts = {}
    try:
        row_counts = {
            i["table_name"]: i["row_count"]
            for i in pipeline.dataset(schema=selected_schema_name)
            .row_counts(dlt_tables=True, load_id=load_id)
            .arrow()
            .to_pylist()
        }
    except (
        DatabaseUndefinedRelation,
        DestinationUndefinedEntity,
        SqlClientNotAvailable,
        PipelineConfigMissing,
        ConnectionError,
    ):
        # TODO: somehow propagate errors to the user here
        pass

    return row_counts


def get_row_counts_list(
    pipeline: dlt.Pipeline, selected_schema_name: str = None, load_id: str = None
) -> List[Dict[str, Union[str, int]]]:
    """Get the row counts for a pipeline as a list of {name, row_count} dicts."""
    row_counts_dict = get_row_counts(pipeline, selected_schema_name, load_id)
    row_counts: List[Dict[str, Union[str, int]]] = [
        {"name": k, "row_count": v} for k, v in row_counts_dict.items()
    ]
    row_counts.sort(key=lambda x: str(x["name"]))
    return row_counts


def get_loads(
    c: DashboardConfiguration, pipeline: dlt.Pipeline, limit: int = 100
) -> Tuple[List[TLoadItem], Optional[str], Optional[str]]:
    """Get the loads of a pipeline. Returns (loads_list, error_message, traceback)."""
    try:
        loads = (
            pipeline.dataset()
            ._dlt_loads.filter("schema_name", "in", pipeline.schema_names)
            .order_by("inserted_at", "desc")
        )
        if limit:
            loads = loads.limit(limit)

        loads_list = loads.arrow().to_pylist()
        loads_list = [humanize_datetime_values(c, load) for load in loads_list]
        return loads_list, None, None
    except Exception as exc:
        return [], format_exception_message(exc), traceback.format_exc()


def build_load_details(
    pipeline: dlt.Pipeline,
    load: TLoadItem,
) -> List[mo.Html]:
    """Build the load detail widgets: row counts and schema version accordion."""
    from dlt._workspace.helpers.dashboard.utils import ui

    result: List[mo.Html] = [mo.md(strings.loads_details_title.format(load["load_id"]))]

    with mo.status.spinner(title=strings.loads_details_loading_spinner_text):
        _schema = get_schema_by_version(pipeline, load["schema_version_hash"])
        _row_counts = get_row_counts_list(pipeline, load["schema_name"], load["load_id"])

    result.append(
        ui.title_and_subtitle(
            strings.loads_details_row_counts_title,
            strings.loads_details_row_counts_subtitle,
            3,
        ),
    )
    result.append(ui.dlt_table(_row_counts))

    if _schema:
        result.append(
            ui.title_and_subtitle(
                strings.loads_details_schema_version_title,
                strings.loads_details_schema_version_subtitle.format(
                    (
                        "is not"
                        if _schema.version_hash != pipeline.default_schema.version_hash
                        else "is"
                    ),
                ),
                3,
            )
        )
        result.append(
            mo.accordion(
                {
                    strings.schema_show_raw_yaml_text: mo.ui.code_editor(
                        _schema.to_pretty_yaml(),
                        language="yaml",
                    )
                }
            )
        )

    return result



from datetime import datetime  # noqa: I251
from typing import Any, Optional, Set, Tuple, List, Type
from pendulum.tz import UTC
from pendulum import DateTime  # noqa: I251

from dlt.common import logger
from dlt.common.exceptions import MissingDependencyException
from dlt.common.utils import digest128
from dlt.common.json import json
from dlt.common.pendulum import create_dt, pendulum
from dlt.common.typing import TDataItem, TColumnNames
from dlt.common.jsonpath import find_values, compile_path, extract_simple_field_name
from dlt.extract.incremental.exceptions import (
    IncrementalCursorInvalidCoercion,
    IncrementalCursorPathMissing,
    IncrementalPrimaryKeyMissing,
    IncrementalCursorPathHasValueNone,
)
from dlt.common.incremental.typing import (
    TCursorValue,
    LastValueFunc,
    OnCursorValueMissing,
    TIncrementalRange,
)
from dlt.extract.utils import resolve_column_value
from dlt.extract.items import TTableHintTemplate

try:
    from dlt.common.libs import pyarrow
    from dlt.common.libs.pyarrow import pyarrow as pa, TAnyArrowItem
    from dlt.common.libs.pyarrow import from_arrow_scalar, to_arrow_scalar
except MissingDependencyException:
    pa = None
    pyarrow = None

try:
    from dlt.common.libs.numpy import numpy
except MissingDependencyException:
    numpy = None

# NOTE: always import pandas independently from pyarrow
try:
    from dlt.common.libs.pandas import pandas, pandas_to_arrow
except MissingDependencyException:
    pandas = None


class IncrementalTransform:
    """A base class for handling extraction and stateful tracking
    of incremental data from input data items.

    By default, the descendant classes are instantiated within the
    `dlt.extract.incremental.Incremental` class.

    Subclasses must implement the `__call__` method which will be called
    for each data item in the extracted data.
    """

    def __init__(
        self,
        resource_name: str,
        cursor_path: str,
        initial_value: Optional[TCursorValue],
        start_value: Optional[TCursorValue],
        end_value: Optional[TCursorValue],
        last_value_func: LastValueFunc[TCursorValue],
        primary_key: Optional[TTableHintTemplate[TColumnNames]],
        unique_hashes: Set[str],
        on_cursor_value_missing: OnCursorValueMissing = "raise",
        lag: Optional[float] = None,
        range_start: TIncrementalRange = "closed",
        range_end: TIncrementalRange = "open",
    ) -> None:
        self.resource_name = resource_name
        self.cursor_path = cursor_path
        self.initial_value = initial_value
        self.start_value = start_value
        self.start_value_is_datetime = isinstance(start_value, datetime)
        # self.start_value_tz = start_value.tzinfo if isinstance(start_value, datetime) else None
        self.last_value = start_value
        self.end_value = end_value
        self.end_value_is_datetime = isinstance(end_value, datetime)
        # self.end_value = end_value.tzinfo if isinstance(end_value, datetime) else None
        self.last_rows: List[TDataItem] = []
        self.last_value_func = last_value_func
        self.unique_hashes = unique_hashes
        self.start_unique_hashes = set(unique_hashes)
        self.on_cursor_value_missing = on_cursor_value_missing
        self.lag = lag
        self.range_start = range_start
        self.range_end = range_end
        # NOTE: self.primary_key is a property
        self.primary_key = primary_key
        # tells if incremental instance has processed any data
        self.seen_data = False

        # compile jsonpath
        self._compiled_cursor_path = compile_path(cursor_path)
        # for simple column name we'll fallback to search in dict
        if simple_field_name := extract_simple_field_name(self._compiled_cursor_path):
            self.cursor_path = simple_field_name
            self._compiled_cursor_path = None

    def compute_unique_value(
        self,
        row: TDataItem,
        primary_key: Optional[TTableHintTemplate[TColumnNames]],
    ) -> str:
        try:
            if primary_key:
                return digest128(json.dumps(resolve_column_value(primary_key, row), sort_keys=True))
            elif primary_key is None:
                return digest128(json.dumps(row, sort_keys=True))
            else:
                return None
        except KeyError as k_err:
            raise IncrementalPrimaryKeyMissing(self.resource_name, k_err.args[0], row)

    def __call__(
        self,
        row: TDataItem,
    ) -> Tuple[bool, bool, bool]: ...

    @staticmethod
    def _adapt_timezone(
        row_value: datetime, cursor_value: datetime, cursor_value_name: str, resource_name: str
    ) -> Any:
        """Adapt cursor_value tz-awareness to match row_value. This method is called only once in instance lifecycle
        to catch incompatible initial values or adapt to change of tz-awareness in the data itself.
        Returns: adjusted cursor_value (with TZ added or removed)
        """
        if (
            cursor_value.tzinfo is None or row_value.tzinfo is None
        ) and cursor_value.tzinfo != row_value.tzinfo:
            config_value_name = (
                "initial_value" if cursor_value_name == "last_value" else cursor_value_name
            )
            last_value_source_message = ""
            if cursor_value_name == "last_value":
                last_value_source_message = (
                    f"{cursor_value_name} is coming from the {config_value_name} defined on"
                    " Incremental or from previous run state. "
                )
            message = (
                f"In resource {resource_name}: {cursor_value_name} '{cursor_value}' and row value"
                f" {row_value} have different timezone awareness. {last_value_source_message}Row"
                f" value is the actual data. {cursor_value_name} will be corrected to match the row"
                " value timezone. The reason for that may be: (1) you set wrong tz-awareness on"
                f" the {config_value_name} (note that pendulum is tz-aware by default) (2) the data"
                " has changed its tz-awareness across runs. (3) your pipeline state got upgraded"
                " to always store last value with tz-awareness following your data."
            )
            logger.warning(message)
            if row_value.tzinfo is None:
                cursor_value = (
                    create_dt(
                        cursor_value.year,
                        cursor_value.month,
                        cursor_value.day,
                        cursor_value.hour,
                        cursor_value.minute,
                        cursor_value.second,
                        cursor_value.microsecond,
                        tz=cursor_value.tzinfo,
                        fold=cursor_value.fold,
                    )
                    .in_tz(tz=UTC)
                    .naive()
                )
            else:
                cursor_value = create_dt(
                    cursor_value.year,
                    cursor_value.month,
                    cursor_value.day,
                    cursor_value.hour,
                    cursor_value.minute,
                    cursor_value.second,
                    cursor_value.microsecond,
                    tz=cursor_value.tzinfo,
                    fold=cursor_value.fold,
                ).in_tz(
                    row_value.tzinfo  # type: ignore[arg-type]
                )
        return cursor_value

    def compute_deduplication_disabled(self) -> bool:
        """Skip deduplication when length of the key is 0 or if lag is applied."""
        # disable deduplication if end value is set - state is not saved
        if self.range_start == "open":
            return True
        if self.end_value is not None:
            return True
        # disable deduplication if lag is applied - destination must deduplicate ranges
        if self.lag and self.last_value_func in (min, max):
            return True
        # disable deduplication if primary_key = ()
        return isinstance(self._primary_key, (list, tuple)) and len(self._primary_key) == 0

    @property
    def primary_key(self) -> Optional[TTableHintTemplate[TColumnNames]]:
        return self._primary_key

    @primary_key.setter
    def primary_key(self, value: str) -> None:
        self._primary_key = value
        self.boundary_deduplication = not self.compute_deduplication_disabled()


class JsonIncremental(IncrementalTransform):
    """Extracts incremental data from JSON data items."""

    def find_cursor_value(self, row: TDataItem) -> Any:
        """Finds value in row at cursor defined by self.cursor_path.

        Will use compiled JSONPath if present.
        Otherwise, reverts to field access if row is dict, Pydantic model, or of other class.
        """
        key_exc: Type[Exception] = IncrementalCursorPathHasValueNone
        if self._compiled_cursor_path:
            # ignores the other found values, e.g. when the path is $data.items[*].created_at
            try:
                row_value = find_values(self._compiled_cursor_path, row)[0]
            except IndexError:
                # empty list so raise a proper exception
                row_value = None
                key_exc = IncrementalCursorPathMissing
        else:
            try:
                try:
                    row_value = row[self.cursor_path]
                except TypeError:
                    # supports Pydantic models and other classes
                    row_value = getattr(row, self.cursor_path)
            except (KeyError, AttributeError):
                # attr not found so raise a proper exception
                row_value = None
                key_exc = IncrementalCursorPathMissing

        # if we have a value - return it
        if row_value is not None:
            return row_value

        if self.on_cursor_value_missing == "raise":
            # raise missing path or None value exception
            raise key_exc(self.resource_name, self.cursor_path, row)
        elif self.on_cursor_value_missing == "exclude":
            return None

    def __call__(
        self,
        row: TDataItem,
    ) -> Tuple[Optional[TDataItem], bool, bool]:
        """
        Returns:
            Tuple (row, start_out_of_range, end_out_of_range) where row is either the data item or `None` if it is completely filtered out
        """
        if row is None:
            return row, False, False

        row_value = self.find_cursor_value(row)
        if row_value is None:
            if self.on_cursor_value_missing == "exclude":
                return None, False, False
            else:
                return row, False, False
        last_value = self.last_value
        last_value_func = self.last_value_func
        # correct tz-awareness if last_value/start_value not matching
        if not self.seen_data and self.start_value_is_datetime and isinstance(row_value, datetime):
            # NOTE: we are making sure that last_value == start_value here (self.seen_data)
            assert self.last_value == self.start_value
            self.last_value = self.start_value = last_value = self._adapt_timezone(
                row_value, last_value, "last_value", self.resource_name
            )

        # Check whether end_value has been reached
        # Filter end value ranges exclusively, so in case of "max" function we remove values >= end_value
        if self.end_value is not None:
            # correct tz-awareness of end value
            if (
                not self.seen_data
                and self.end_value_is_datetime
                and isinstance(row_value, datetime)
            ):
                self.end_value = self._adapt_timezone(
                    row_value, self.end_value, "end_value", self.resource_name
                )
            try:
                if last_value_func((row_value, self.end_value)) != self.end_value:
                    return None, False, True

                if self.range_end == "open" and last_value_func((row_value,)) == self.end_value:
                    return None, False, True
            except Exception as ex:
                raise IncrementalCursorInvalidCoercion(
                    self.resource_name,
                    self.cursor_path,
                    self.end_value,
                    "end_value",
                    row_value,
                    type(row_value).__name__,
                    str(ex),
                ) from ex
        # row was read
        self.seen_data = True

        if last_value is None:
            check_values: Tuple[Any, ...] = (row_value,)
        else:
            check_values = (row_value, last_value)

        try:
            new_value = last_value_func(check_values)
        except Exception as ex:
            raise IncrementalCursorInvalidCoercion(
                self.resource_name,
                self.cursor_path,
                last_value,
                "start_value/initial_value",
                row_value,
                type(row_value).__name__,
                str(ex),
            ) from ex
        # new_value is "less" or equal to last_value (the actual max)
        if last_value == new_value:
            # use func to compute row_value into last_value compatible
            processed_row_value = last_value_func((row_value,))

            if self.range_start == "open" and processed_row_value == self.start_value:
                # We only want greater than start_value
                return None, True, False

            # skip the record that is not a start_value or new_value: that record was already processed
            if self.start_value is None:
                check_values = (row_value,)
            else:
                check_values = (row_value, self.start_value)
            new_value = last_value_func(check_values)
            # Include rows == start_value but exclude "lower"
            # new_value is "less" or equal to start_value (the initial max)
            if new_value == self.start_value:
                # if equal there's still a chance that item gets in
                if processed_row_value == self.start_value:
                    if self.boundary_deduplication:
                        unique_value = self.compute_unique_value(row, self._primary_key)
                        # if unique value exists then use it to deduplicate
                        if unique_value in self.start_unique_hashes:
                            return None, True, False
                else:
                    # "smaller" than start value: gets out
                    return None, True, False

            # we store row id for all records with the current "last_value" in state and use it to deduplicate
            if processed_row_value == last_value and self.boundary_deduplication:
                # add new hash only if the record row id is same as current last value
                self.last_rows.append(row)
        else:
            self.last_value = new_value
            if self.boundary_deduplication:
                # store rows with "max" values to compute hashes after processing full batch
                self.last_rows = [row]
                self.unique_hashes = set()

        return row, False, False


class ArrowIncremental(IncrementalTransform):
    _dlt_index = "_dlt_index"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        if self.last_value_func is max:
            self.compute = pa.compute.max
            self.end_compare = (
                pa.compute.less if self.range_end == "open" else pa.compute.less_equal
            )
            self.last_value_compare = (
                pa.compute.greater_equal if self.range_start == "closed" else pa.compute.greater
            )
            self.new_value_compare = pa.compute.greater
        elif self.last_value_func is min:
            self.compute = pa.compute.min
            self.end_compare = (
                pa.compute.greater if self.range_end == "open" else pa.compute.greater_equal
            )
            self.last_value_compare = (
                pa.compute.less_equal if self.range_start == "closed" else pa.compute.less
            )
            self.new_value_compare = pa.compute.less
        else:
            raise NotImplementedError(
                "Only `min` or `max` of `last_value_func` is supported for arrow tables"
            )

    def compute_unique_values(self, item: "TAnyArrowItem", unique_columns: List[str]) -> List[str]:
        if not unique_columns:
            return []
        rows = item.select(unique_columns).to_pylist()
        return [self.compute_unique_value(row, self._primary_key) for row in rows]

    def compute_unique_values_with_index(
        self, item: "TAnyArrowItem", unique_columns: List[str]
    ) -> List[Tuple[Any, str]]:
        if not unique_columns:
            return []
        indices = item[self._dlt_index].to_pylist()
        rows = item.select(unique_columns).to_pylist()
        return [
            (index, self.compute_unique_value(row, self._primary_key))
            for index, row in zip(indices, rows)
        ]

    def _add_unique_index(self, tbl: "pa.Table") -> "pa.Table":
        """Creates unique index if necessary."""
        # create unique index if necessary
        if self._dlt_index not in tbl.schema.names:
            # indices = pa.compute.sequence(start=0, step=1, length=tbl.num_rows,
            indices = pa.array(range(tbl.num_rows))
            tbl = pyarrow.append_column(tbl, self._dlt_index, indices)
        return tbl

    def __call__(
        self,
        tbl: "TAnyArrowItem",
    ) -> Tuple[TDataItem, bool, bool]:
        is_pandas = pandas is not None and isinstance(tbl, pandas.DataFrame)
        if is_pandas:
            tbl = pandas_to_arrow(tbl)

        primary_key = self._primary_key(tbl) if callable(self._primary_key) else self._primary_key
        if primary_key:
            # create a list of unique columns
            if isinstance(primary_key, str):
                unique_columns = [primary_key]
            else:
                unique_columns = list(primary_key)
            # check if primary key components are in the table
            for pk in unique_columns:
                if pk not in tbl.schema.names:
                    raise IncrementalPrimaryKeyMissing(self.resource_name, pk, tbl)
            # use primary key as unique index
            if isinstance(primary_key, str):
                self._dlt_index = primary_key
        elif primary_key is None:
            unique_columns = tbl.schema.names

        start_out_of_range = end_out_of_range = False
        if not tbl:  # row is None or empty arrow table
            return tbl, start_out_of_range, end_out_of_range

        # TODO: Json path support. For now assume the cursor_path is a column name
        cursor_path = self.cursor_path
        try:
            cursor_data_type = tbl.schema.field(cursor_path).type
        except KeyError as e:
            raise IncrementalCursorPathMissing(
                self.resource_name,
                cursor_path,
                tbl,
                f"Column name `{cursor_path}` was not found in the arrow table. Nested JSON paths"
                " are not supported for arrow tables and dataframes, the incremental `cursor_path`"
                " must be a column name.",
            ) from e

        # The new max/min value
        row_value_scalar = self.compute(tbl[cursor_path])
        row_value = from_arrow_scalar(row_value_scalar)
        # correct tz-awareness
        if not self.seen_data and self.start_value_is_datetime and isinstance(row_value, datetime):
            # NOTE: we are making sure that last_value == start_value here (self.seen_data)
            assert self.last_value == self.start_value
            self.start_value = self.last_value = self._adapt_timezone(
                row_value, self.last_value, "last_value", self.resource_name
            )

        if tbl.schema.field(cursor_path).nullable:
            tbl_without_null, tbl_with_null = self._process_null_at_cursor_path(tbl)
            tbl = tbl_without_null

        # If end_value is provided, filter to include table rows that are "less" than end_value
        if self.end_value is not None:
            # correct tz-awareness of end value
            if (
                not self.seen_data
                and self.end_value_is_datetime
                and isinstance(row_value, datetime)
            ):
                self.end_value = self._adapt_timezone(
                    row_value, self.end_value, "end_value", self.resource_name
                )
            try:
                end_value_scalar = to_arrow_scalar(self.end_value, cursor_data_type)
            except Exception as ex:
                raise IncrementalCursorInvalidCoercion(
                    self.resource_name,
                    cursor_path,
                    self.end_value,
                    "end_value",
                    "<arrow column>",
                    cursor_data_type,
                    str(ex),
                ) from ex
            # Is max row value higher than end value?
            # NOTE: pyarrow bool *always* evaluates to python True. `as_py()` is necessary
            end_out_of_range = not self.end_compare(row_value_scalar, end_value_scalar).as_py()
            if end_out_of_range:
                tbl = tbl.filter(self.end_compare(tbl[cursor_path], end_value_scalar))
                # NOTE: here we should recalculate row_value_scalar and row_value because it was out of range!
                # right now we don't do that because self.last_value is not saved into state so this value will
                # be discarded anyway. If we ever start saving state when end_value is present then we must able that
                # row_value_scalar = self.compute(
                #     tbl[cursor_path]
                # )
                # row_value = from_arrow_scalar(row_value_scalar)

        self.seen_data = True

        if self.start_value is not None:
            try:
                start_value_scalar = to_arrow_scalar(self.start_value, cursor_data_type)
            except Exception as ex:
                raise IncrementalCursorInvalidCoercion(
                    self.resource_name,
                    cursor_path,
                    self.start_value,
                    "start_value/initial_value",
                    "<arrow column>",
                    cursor_data_type,
                    str(ex),
                ) from ex
            # Remove rows lower or equal than the last start value
            keep_filter = self.last_value_compare(tbl[cursor_path], start_value_scalar)
            start_out_of_range = bool(pa.compute.any(pa.compute.invert(keep_filter)).as_py())
            tbl = tbl.filter(keep_filter)
            if self.boundary_deduplication:
                # Deduplicate after filtering old values
                tbl = self._add_unique_index(tbl)
                # Remove already processed rows where the cursor is equal to the start value
                eq_rows = tbl.filter(pa.compute.equal(tbl[cursor_path], start_value_scalar))
                # compute index, unique hash mapping
                unique_values_index = self.compute_unique_values_with_index(eq_rows, unique_columns)
                unique_values_index = [
                    (i, uq_val)
                    for i, uq_val in unique_values_index
                    if uq_val in self.start_unique_hashes
                ]
                if len(unique_values_index) > 0:
                    # find rows with unique ids that were stored from previous run
                    remove_idx = pa.array(i for i, _ in unique_values_index)
                    tbl = tbl.filter(
                        pa.compute.invert(pa.compute.is_in(tbl[self._dlt_index], remove_idx))
                    )

        if (
            self.last_value is None
            or self.new_value_compare(
                row_value_scalar, to_arrow_scalar(self.last_value, cursor_data_type)
            ).as_py()
        ):  # Last value has changed
            self.last_value = row_value
            if self.boundary_deduplication:
                # Compute unique hashes for all rows equal to row value
                self.unique_hashes = set(
                    self.compute_unique_values(
                        tbl.filter(pa.compute.equal(tbl[cursor_path], row_value_scalar)),
                        unique_columns,
                    )
                )
        elif self.last_value == row_value and self.boundary_deduplication:
            # last value is unchanged, add the hashes
            self.unique_hashes.update(
                set(
                    self.compute_unique_values(
                        tbl.filter(pa.compute.equal(tbl[cursor_path], row_value_scalar)),
                        unique_columns,
                    )
                )
            )

        # drop the temp unique index before concat and returning
        if "_dlt_index" in tbl.schema.names:
            tbl = pyarrow.remove_columns(tbl, ["_dlt_index"])

        if self.on_cursor_value_missing == "include":
            if tbl.schema.field(cursor_path).nullable:
                if isinstance(tbl, pa.RecordBatch):
                    assert isinstance(tbl_with_null, pa.RecordBatch)
                    tbl = pa.Table.from_batches([tbl, tbl_with_null])
                else:
                    tbl = pa.concat_tables([tbl, tbl_with_null], promote_options="none")

        if len(tbl) == 0:
            return None, start_out_of_range, end_out_of_range
        if is_pandas:
            tbl = tbl.to_pandas()
        return tbl, start_out_of_range, end_out_of_range

    def _process_null_at_cursor_path(self, tbl: "pa.Table") -> Tuple["pa.Table", "pa.Table"]:
        mask = pa.compute.is_valid(tbl[self.cursor_path])
        rows_without_null = tbl.filter(mask)
        rows_with_null = tbl.filter(pa.compute.invert(mask))
        if self.on_cursor_value_missing == "raise":
            if rows_with_null.num_rows > 0:
                raise IncrementalCursorPathHasValueNone(self.resource_name, self.cursor_path)
        return rows_without_null, rows_with_null



from typing import ClassVar, Literal, Optional

from dlt.common.configuration import configspec, known_sections
from dlt.common.configuration.specs import BaseConfiguration
from dlt.common.time import get_precision_from_datetime_unit

CsvQuoting = Literal["quote_all", "quote_needed", "quote_minimal", "quote_none"]


@configspec
class CsvFormatConfiguration(BaseConfiguration):
    delimiter: str = ","
    include_header: bool = True
    quoting: CsvQuoting = "quote_needed"
    lineterminator: str = "\n"

    # read options
    on_error_continue: bool = False
    encoding: str = "utf-8"

    __section__: ClassVar[str] = known_sections.DATA_WRITER


@configspec
class ParquetFormatConfiguration(BaseConfiguration):
    flavor: Optional[str] = None  # could be ie. "spark"
    version: Optional[str] = "2.4"
    data_page_size: Optional[int] = None
    timestamp_timezone: str = "UTC"
    row_group_size: Optional[int] = None
    coerce_timestamps: Optional[Literal["s", "ms", "us", "ns"]] = None
    allow_truncated_timestamps: bool = False
    use_compliant_nested_type: bool = True
    write_page_index: bool = False
    use_content_defined_chunking: bool = False  # requires pyarrow>=21.0.0, ignored otherwise
    supports_dictionary_encoding: bool = True
    """When False, constant columns (like _dlt_load_id) will use regular arrays instead of
    dictionary-encoded arrays. Set to False for destinations using ADBC drivers that don't
    support dictionary types (e.g., MSSQL)."""

    def max_timestamp_precision(self) -> int:
        if (self.flavor or "").lower() == "spark":
            base = get_precision_from_datetime_unit("ns")  # INT96 -> treat as ns-capable
        else:
            v = float(self.version or "0.0")
            base = (
                get_precision_from_datetime_unit("ns")
                if v >= 2.6
                else get_precision_from_datetime_unit("us")
            )

        if self.coerce_timestamps:
            return min(base, get_precision_from_datetime_unit(self.coerce_timestamps))
        return base

    __section__: ClassVar[str] = known_sections.DATA_WRITER



from typing import Any, Dict, List
import requests

from dlt.common.configuration import with_config

from dlt._workspace.cli.exceptions import AiContextApiError
from dlt._workspace.configuration import WorkspaceRuntimeConfiguration
from dlt._workspace.typing import TSourceItem


# Timeout for AI context API requests (20 seconds to handle larger files)
AI_CONTEXT_API_TIMEOUT = 20


@with_config(spec=WorkspaceRuntimeConfiguration, sections=("runtime", "workspace"))
def search_sources(query: str = "", ai_context_api_url: str = None) -> List[TSourceItem]:
    """Search for available sources by name and description.

    Args:
        query: Search string. Empty string returns all sources alphabetically.
        ai_context_api_url: Base URL of the AI context API (auto-injected from config)

    Returns:
        List of source items, each with source_name, description, and optional
        description_verbose and sample_urls fields.

    Raises:
        AiContextApiError: If there's an error connecting to the API
    """
    url = f"{ai_context_api_url}/api/v1/scaffolds/sources"
    params = {"q": query} if query else {}
    try:
        response = requests.get(url, params=params, timeout=AI_CONTEXT_API_TIMEOUT)
    except requests.RequestException as e:
        raise AiContextApiError(f"There was an error connecting to the AI context API: {str(e)}")

    if response.status_code != 200:
        raise AiContextApiError(f"API returned status {response.status_code}: {response.text}")

    data: Dict[str, Any] = response.json()
    results: List[TSourceItem] = data.get("results", [])
    return results



from contextlib import contextmanager, suppress
from typing import Any, AnyStr, ClassVar, Iterator, Optional, Sequence, List, Tuple

import pyarrow

from dlt.common import logger
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.destinations.exceptions import (
    DatabaseTerminalException,
    DatabaseUndefinedRelation,
    DatabaseTransientException,
)
from dlt.destinations.impl.dremio import pydremio
from dlt.destinations.impl.dremio.configuration import DremioCredentials
from dlt.destinations.sql_client import (
    DBApiCursorImpl,
    SqlClientBase,
    raise_database_error,
    raise_open_connection_error,
)
from dlt.destinations.typing import DBApi, DBTransaction, DataFrame
from dlt.common.destination.dataset import DBApiCursor


class DremioCursorImpl(DBApiCursorImpl):
    native_cursor: pydremio.DremioCursor  # type: ignore[assignment]

    def df(self, chunk_size: int = None, **kwargs: Any) -> Optional[DataFrame]:
        if chunk_size is None:
            return self.arrow(chunk_size=chunk_size).to_pandas()
        return super().df(chunk_size=chunk_size, **kwargs)

    def arrow(self, chunk_size: int = None, **kwargs: Any) -> Optional[DataFrame]:
        if chunk_size is None:
            return self.native_cursor.fetch_arrow_table()
        return self.native_cursor.iter_arrow_tables(chunk_size)


class DremioSqlClient(SqlClientBase[pydremio.DremioConnection]):
    dbapi: ClassVar[DBApi] = pydremio
    SENTINEL_TABLE_NAME: ClassVar[str] = "_dlt_sentinel_table"

    def __init__(
        self,
        dataset_name: str,
        staging_dataset_name: str,
        credentials: DremioCredentials,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        super().__init__(credentials.database, dataset_name, staging_dataset_name, capabilities)
        self._conn: Optional[pydremio.DremioConnection] = None
        self.credentials = credentials

    def open_connection(self) -> pydremio.DremioConnection:
        db_kwargs = self.credentials.db_kwargs()
        self._conn = pydremio.connect(
            uri=self.credentials.to_native_credentials(), db_kwargs=db_kwargs
        )
        return self._conn

    @raise_open_connection_error
    def close_connection(self) -> None:
        if self._conn:
            self._conn.close()
            self._conn = None

    @contextmanager
    @raise_database_error
    def begin_transaction(self) -> Iterator[DBTransaction]:
        logger.warning(
            "Dremio does not support transactions! Each SQL statement is auto-committed separately."
        )
        yield self

    @raise_database_error
    def commit_transaction(self) -> None:
        pass

    @raise_database_error
    def rollback_transaction(self) -> None:
        raise NotImplementedError("You cannot rollback Dremio SQL statements.")

    @property
    def native_connection(self) -> "pydremio.DremioConnection":
        return self._conn

    def drop_tables(self, *tables: str) -> None:
        # Tables are drop with `IF EXISTS`, but dremio raises when the schema doesn't exist.
        # Multi statement exec is safe and the error can be ignored since all tables are in the same schema.
        with suppress(DatabaseUndefinedRelation):
            super().drop_tables(*tables)

    def execute_sql(
        self, sql: AnyStr, *args: Any, **kwargs: Any
    ) -> Optional[Sequence[Sequence[Any]]]:
        with self.execute_query(sql, *args, **kwargs) as curr:
            if curr.description is None:
                return None
            else:
                f = curr.fetchall()
                return f

    @contextmanager
    @raise_database_error
    def execute_query(self, query: AnyStr, *args: Any, **kwargs: Any) -> Iterator[DBApiCursor]:
        db_args = args if args else kwargs if kwargs else None
        with self._conn.cursor() as curr:
            try:
                curr.execute(query, db_args)  # type: ignore
            except pydremio.MalformedQueryError as ex:
                raise DatabaseTransientException(ex)
            yield DremioCursorImpl(curr)  # type: ignore

    def catalog_name(self, quote: bool = True, casefold: bool = True) -> Optional[str]:
        if casefold:
            database_name = self.capabilities.casefold_identifier(self.database_name)
        else:
            database_name = self.database_name
        if quote:
            database_name = self.capabilities.escape_identifier(database_name)
        return database_name

    def _get_information_schema_components(self, *tables: str) -> Tuple[str, str, List[str]]:
        components = super()._get_information_schema_components(*tables)
        # catalog is always DREMIO but schema contains "database" prefix 🤷
        return ("DREMIO", self.fully_qualified_dataset_name(quote=False), components[2])

    @classmethod
    def _make_database_exception(cls, ex: Exception) -> Exception:
        if isinstance(ex, pyarrow.lib.ArrowInvalid):
            msg = str(ex)
            if "not found in any table" in msg:
                return DatabaseTerminalException(ex)
            elif "not found" in msg or "does not exist" in msg:
                return DatabaseUndefinedRelation(ex)
            elif "Non-query expression encountered in illegal context" in msg:
                return DatabaseTransientException(ex)
            else:
                return DatabaseTerminalException(ex)
        else:
            return ex

    @staticmethod
    def is_dbapi_exception(ex: Exception) -> bool:
        return isinstance(ex, (pyarrow.lib.ArrowInvalid, pydremio.MalformedQueryError))

    def create_dataset(self) -> None:
        # We create a sentinel table which defines wether we consider the dataset created
        sentinel_table_name = self.make_qualified_table_name(self.SENTINEL_TABLE_NAME)
        self.execute_sql(f"CREATE TABLE {sentinel_table_name} (_dlt_id BIGINT)")

    def _get_table_names(self) -> List[str]:
        query = """
            SELECT TABLE_NAME
            FROM INFORMATION_SCHEMA."TABLES"
            WHERE TABLE_CATALOG = %s AND TABLE_SCHEMA = %s
            """
        catalog_name, schema_name, _ = self._get_information_schema_components()
        tables = self.execute_sql(query, catalog_name, schema_name) or []
        return [table[0] for table in tables]

    def drop_dataset(self) -> None:
        # drop sentinel table
        sentinel_table_name = self.make_qualified_table_name(self.SENTINEL_TABLE_NAME)
        # must exist or we get undefined relation exception
        self.execute_sql(f"DROP TABLE {sentinel_table_name}")

        table_names = self._get_table_names()
        for table_name in table_names:
            full_table_name = self.make_qualified_table_name(table_name)
            self.execute_sql("DROP TABLE IF EXISTS %s" % full_table_name)

    def has_dataset(self) -> bool:
        return len(self._get_table_names()) > 0



"""Schema introspection helpers: table lists, column lists, schema retrieval, and resource state."""

from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from dlt._workspace.helpers.dashboard.typing import TNameValueItem, TTableListItem

import marimo as mo

import dlt

from dlt.common.destination.client import WithStateSync
from dlt.common.json import json
from dlt.common.schema import Schema
from dlt.common.schema.typing import TTableSchema
from dlt.common.typing import DictStrAny

from dlt._workspace.helpers.dashboard import strings
from dlt._workspace.helpers.dashboard.config import DashboardConfiguration
from dlt._workspace.helpers.dashboard.utils.ui import BoundedDict, small


def schemas_to_table_items(
    schemas: Iterable[Schema], default_schema_name: str
) -> List[TNameValueItem]:
    """Convert a list of schemas to name/value display items, with default schema first."""
    schemas = sorted(
        schemas, key=lambda s: 0 if getattr(s, "name", None) == default_schema_name else 1
    )
    table_items: List[TNameValueItem] = []
    count = 0
    for schema in schemas:
        table_items.append(
            {
                "name": "schemas" if count == 0 else "",
                "value": (
                    schema.name
                    + f" ({schema.version}, {schema.version_hash[:8]}) "
                    + (" (default)" if schema.name == default_schema_name else "")
                ),
            }
        )
        count += 1
    return table_items


def create_table_list(
    c: DashboardConfiguration,
    pipeline: dlt.Pipeline,
    selected_schema_name: str = None,
    show_internals: bool = False,
    show_child_tables: bool = True,
    show_row_counts: bool = False,
) -> List[TTableListItem]:
    """Create a list of tables for the pipeline, optionally including internals, child tables, and row counts."""
    from dlt._workspace.helpers.dashboard.utils.queries import get_row_counts

    tables = list(
        pipeline.schemas[selected_schema_name].data_tables(
            seen_data_only=True, include_incomplete=False
        )
    )
    if not show_child_tables:
        tables = [t for t in tables if t.get("parent") is None]

    if show_internals:
        tables = tables + list(pipeline.schemas[selected_schema_name].dlt_tables())

    row_counts = get_row_counts(pipeline, selected_schema_name) if show_row_counts else {}
    table_list: List[TTableListItem] = []
    for table in tables:
        item = TTableListItem(
            name=table.get("name", ""),
            row_count=row_counts.get(table["name"], None),
        )
        for prop in c.table_list_fields:
            item[prop] = table.get(prop, None)  # type: ignore[literal-required]
        table_list.append(item)
    table_list.sort(key=lambda x: x["name"])
    return table_list


def create_column_list(
    c: DashboardConfiguration,
    pipeline: dlt.Pipeline,
    table_name: str,
    selected_schema_name: str = None,
    show_internals: bool = False,
    show_type_hints: bool = True,
    show_other_hints: bool = False,
    show_custom_hints: bool = False,
) -> List[Dict[str, Any]]:
    """Create a list of columns for a table, with configurable hint visibility."""
    column_list: List[Dict[str, Any]] = []
    for column in (
        pipeline.schemas[selected_schema_name]
        .get_table_columns(table_name, include_incomplete=False)
        .values()
    ):
        column_dict: Dict[str, Any] = {
            "name": column["name"],
        }

        if show_type_hints:
            column_dict = {
                **column_dict,
                **{hint: column.get(hint, None) for hint in c.column_type_hints},
            }

        if show_other_hints:
            column_dict = {
                **column_dict,
                **{hint: column.get(hint, None) for hint in c.column_other_hints},
            }

        if show_custom_hints:
            for key in column:
                if key.startswith("x-"):
                    column_dict[key] = column[key]  # type: ignore

        column_list.append(column_dict)

    if not show_internals:
        column_list = [c for c in column_list if not c["name"].lower().startswith("_dlt")]
    return column_list


def get_source_and_resource_state_for_table(
    table: TTableSchema, pipeline: dlt.Pipeline, schema_name: str
) -> Tuple[str, DictStrAny, DictStrAny]:
    """Return (resource_name, source_state, resource_state) for the resource that created the given table."""
    if "resource" not in table:
        return None, {}, {}

    pipeline.activate()
    resource_name = table["resource"]
    source_state = dlt.extract.state.source_state(schema_name)
    resource_state = dlt.extract.state.resource_state(resource_name, source_state)
    source_state = {k: v for k, v in source_state.items() if k != "resources"}

    return table["resource"], source_state, resource_state


def get_schema_by_version(pipeline: dlt.Pipeline, version_hash: str) -> Schema:
    """Retrieve a schema from the destination by its version hash (cached)."""
    cache_key = (pipeline.pipeline_name, version_hash)
    if cache_key not in _schema_version_cache:
        _schema_version_cache[cache_key] = _fetch_schema_by_version(pipeline, version_hash)
    result: Schema = _schema_version_cache[cache_key]
    return result


_schema_version_cache: BoundedDict = BoundedDict(32)


def _fetch_schema_by_version(pipeline: dlt.Pipeline, version_hash: str) -> Schema:
    """Retrieve a schema from the destination by its version hash."""
    with pipeline.destination_client() as client:
        if isinstance(client, WithStateSync):
            stored_schema = client.get_stored_schema_by_hash(version_hash)
            if not stored_schema:
                return None
            return Schema.from_stored_schema(json.loads(stored_schema.schema))
    return None


def build_resource_state_widget(
    pipeline: dlt.Pipeline, schema_name: str, table_name: str
) -> Optional[mo.Html]:
    """Build source/resource state accordion for the given table, or None if not applicable."""
    schema_table = pipeline.schemas[schema_name].tables[table_name]
    resource_name, source_state, resource_state = get_source_and_resource_state_for_table(
        schema_table, pipeline, schema_name
    )
    if not resource_name:
        return None

    state_content = mo.hstack(
        [
            mo.vstack(
                [
                    mo.md(small(strings.schema_source_state_label.format(schema_name))),
                    mo.json(source_state),
                ]
            ),
            mo.vstack(
                [
                    mo.md(small(strings.schema_resource_state_label.format(resource_name))),
                    mo.json(resource_state),
                ]
            ),
        ],
        justify="start",
        widths="equal",
    )

    return mo.accordion(
        {small(strings.schema_show_resource_state.format(resource_name, table_name)): state_content}
    )
